@@ -3,13 +3,12 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "thread.h"
+#include <mgba/core/thread.h>
 
-#include "core/core.h"
-#include "util/patch.h"
-#include "util/vfs.h"
-
-#include "feature/commandline.h"
+#include <mgba/core/core.h>
+#include <mgba/core/serialize.h>
+#include <mgba-util/patch.h>
+#include <mgba-util/vfs.h>
 
 #include <signal.h>
 
@@ -37,6 +36,8 @@ static BOOL CALLBACK _createTLS(PINIT_ONCE once, PVOID param, PVOID* context) {
 }
 #endif
 
+static void _mCoreLog(struct mLogger* logger, int category, enum mLogLevel level, const char* format, va_list args);
+
 static void _changeState(struct mCoreThread* threadContext, enum mCoreThreadState newState, bool broadcast) {
 	MutexLock(&threadContext->stateMutex);
 	threadContext->state = newState;
@@ -85,6 +86,50 @@ static void _pauseThread(struct mCoreThread* threadContext) {
 	_waitUntilNotState(threadContext, THREAD_PAUSING);
 }
 
+void _frameStarted(void* context) {
+	struct mCoreThread* thread = context;
+	if (!thread) {
+		return;
+	}
+	if (thread->core->opts.rewindEnable && thread->core->opts.rewindBufferCapacity > 0) {
+		if (thread->state != THREAD_REWINDING) {
+			mCoreRewindAppend(&thread->rewind, thread->core);
+		} else if (thread->state == THREAD_REWINDING) {
+			if (!mCoreRewindRestore(&thread->rewind, thread->core)) {
+				mCoreRewindAppend(&thread->rewind, thread->core);
+			}
+		}
+	}
+}
+
+void _frameEnded(void* context) {
+	struct mCoreThread* thread = context;
+	if (!thread) {
+		return;
+	}
+	if (thread->frameCallback) {
+		thread->frameCallback(thread);
+	}
+}
+
+void _crashed(void* context) {
+	struct mCoreThread* thread = context;
+	if (!thread) {
+		return;
+	}
+	_changeState(thread, THREAD_CRASHED, true);
+}
+
+void _coreSleep(void* context) {
+	struct mCoreThread* thread = context;
+	if (!thread) {
+		return;
+	}
+	if (thread->sleepCallback) {
+		thread->sleepCallback(thread);
+	}
+}
+
 static THREAD_ENTRY _mCoreThreadRun(void* context) {
 	struct mCoreThread* threadContext = context;
 #ifdef USE_PTHREADS
@@ -104,11 +149,27 @@ static THREAD_ENTRY _mCoreThreadRun(void* context) {
 #endif
 
 	struct mCore* core = threadContext->core;
+	struct mCoreCallbacks callbacks = {
+		.videoFrameStarted = _frameStarted,
+		.videoFrameEnded = _frameEnded,
+		.coreCrashed = _crashed,
+		.sleep = _coreSleep,
+		.context = threadContext
+	};
+	core->addCoreCallbacks(core, &callbacks);
 	core->setSync(core, &threadContext->sync);
 	core->reset(core);
 
+	struct mLogFilter filter;
+	if (!threadContext->logger.d.filter) {
+		threadContext->logger.d.filter = &filter;
+		mLogFilterInit(threadContext->logger.d.filter);
+		mLogFilterLoad(threadContext->logger.d.filter, &core->config);
+	}
+
 	if (core->opts.rewindEnable && core->opts.rewindBufferCapacity > 0) {
-		 mCoreRewindContextInit(&threadContext->rewind, core->opts.rewindBufferCapacity);
+		 mCoreRewindContextInit(&threadContext->rewind, core->opts.rewindBufferCapacity, true);
+		 threadContext->rewind.stateFlags = core->opts.rewindSave ? SAVESTATE_SAVEDATA : 0;
 	}
 
 	_changeState(threadContext, THREAD_RUNNING, true);
@@ -121,13 +182,16 @@ static THREAD_ENTRY _mCoreThreadRun(void* context) {
 	}
 
 	while (threadContext->state < THREAD_EXITING) {
+#ifdef USE_DEBUGGERS
 		struct mDebugger* debugger = core->debugger;
 		if (debugger) {
 			mDebuggerRun(debugger);
 			if (debugger->state == DEBUGGER_SHUTDOWN) {
 				_changeState(threadContext, THREAD_EXITING, false);
 			}
-		} else {
+		} else
+#endif
+		{
 			while (threadContext->state <= THREAD_MAX_RUNNING) {
 				core->runLoop(core);
 			}
@@ -179,6 +243,9 @@ static THREAD_ENTRY _mCoreThreadRun(void* context) {
 	if (threadContext->cleanCallback) {
 		threadContext->cleanCallback(threadContext);
 	}
+	core->clearCoreCallbacks(core);
+
+	threadContext->logger.d.filter = NULL;
 
 	return 0;
 }
@@ -186,7 +253,10 @@ static THREAD_ENTRY _mCoreThreadRun(void* context) {
 bool mCoreThreadStart(struct mCoreThread* threadContext) {
 	threadContext->state = THREAD_INITIALIZED;
 	threadContext->logger.p = threadContext;
-	threadContext->logLevel = threadContext->core->opts.logLevel;
+	if (!threadContext->logger.d.log) {
+		threadContext->logger.d.log = _mCoreLog;
+		threadContext->logger.d.filter = NULL;
+	}
 
 	if (!threadContext->sync.fpsTarget) {
 		threadContext->sync.fpsTarget = _defaultFPSTarget;
@@ -331,11 +401,14 @@ void mCoreThreadInterruptFromThread(struct mCoreThread* threadContext) {
 	MutexLock(&threadContext->stateMutex);
 	++threadContext->interruptDepth;
 	if (threadContext->interruptDepth > 1 || !mCoreThreadIsActive(threadContext)) {
+		if (threadContext->state == THREAD_INTERRUPTING) {
+			threadContext->state = THREAD_INTERRUPTED;
+		}
 		MutexUnlock(&threadContext->stateMutex);
 		return;
 	}
 	threadContext->savedState = threadContext->state;
-	threadContext->state = THREAD_INTERRUPTING;
+	threadContext->state = THREAD_INTERRUPTED;
 	ConditionWake(&threadContext->stateCond);
 	MutexUnlock(&threadContext->stateMutex);
 }
@@ -395,7 +468,7 @@ void mCoreThreadUnpause(struct mCoreThread* threadContext) {
 bool mCoreThreadIsPaused(struct mCoreThread* threadContext) {
 	bool isPaused;
 	MutexLock(&threadContext->stateMutex);
-	if (threadContext->state == THREAD_INTERRUPTED || threadContext->state == THREAD_INTERRUPTING) {
+	if (threadContext->interruptDepth) {
 		isPaused = threadContext->savedState == THREAD_PAUSED;
 	} else {
 		isPaused = threadContext->state == THREAD_PAUSED;
@@ -425,7 +498,7 @@ void mCoreThreadTogglePause(struct mCoreThread* threadContext) {
 void mCoreThreadPauseFromThread(struct mCoreThread* threadContext) {
 	bool frameOn = true;
 	MutexLock(&threadContext->stateMutex);
-	if (threadContext->state == THREAD_RUNNING || (threadContext->state == THREAD_INTERRUPTING && threadContext->savedState == THREAD_RUNNING)) {
+	if (threadContext->state == THREAD_RUNNING || (threadContext->interruptDepth && threadContext->savedState == THREAD_RUNNING)) {
 		threadContext->state = THREAD_PAUSING;
 		frameOn = false;
 	}
@@ -436,6 +509,14 @@ void mCoreThreadPauseFromThread(struct mCoreThread* threadContext) {
 
 void mCoreThreadSetRewinding(struct mCoreThread* threadContext, bool rewinding) {
 	MutexLock(&threadContext->stateMutex);
+	if (rewinding && (threadContext->state == THREAD_REWINDING || (threadContext->interruptDepth && threadContext->savedState == THREAD_REWINDING))) {
+		MutexUnlock(&threadContext->stateMutex);
+		return;
+	}
+	if (!rewinding && ((!threadContext->interruptDepth && threadContext->state != THREAD_REWINDING) || (threadContext->interruptDepth && threadContext->savedState != THREAD_REWINDING))) {
+		MutexUnlock(&threadContext->stateMutex);
+		return;
+	}
 	_waitOnInterrupt(threadContext);
 	if (rewinding && threadContext->state == THREAD_RUNNING) {
 		threadContext->state = THREAD_REWINDING;
@@ -448,7 +529,7 @@ void mCoreThreadSetRewinding(struct mCoreThread* threadContext, bool rewinding)
 
 void mCoreThreadWaitFromThread(struct mCoreThread* threadContext) {
 	MutexLock(&threadContext->stateMutex);
-	if ((threadContext->state == THREAD_INTERRUPTED || threadContext->state == THREAD_INTERRUPTING) && threadContext->savedState == THREAD_RUNNING) {
+	if (threadContext->interruptDepth && threadContext->savedState == THREAD_RUNNING) {
 		threadContext->savedState = THREAD_WAITING;
 	} else if (threadContext->state == THREAD_RUNNING) {
 		threadContext->state = THREAD_WAITING;
@@ -458,7 +539,7 @@ void mCoreThreadWaitFromThread(struct mCoreThread* threadContext) {
 
 void mCoreThreadStopWaiting(struct mCoreThread* threadContext) {
 	MutexLock(&threadContext->stateMutex);
-	if ((threadContext->state == THREAD_INTERRUPTED || threadContext->state == THREAD_INTERRUPTING) && threadContext->savedState == THREAD_WAITING) {
+	if (threadContext->interruptDepth && threadContext->savedState == THREAD_WAITING) {
 		threadContext->savedState = THREAD_RUNNING;
 	} else if (threadContext->state == THREAD_WAITING) {
 		threadContext->state = THREAD_RUNNING;
@@ -483,51 +564,15 @@ struct mCoreThread* mCoreThreadGet(void) {
 }
 #endif
 
-void mCoreThreadFrameStarted(struct mCoreThread* thread) {
-	if (!thread) {
-		return;
-	}
-	if (thread->core->opts.rewindEnable && thread->core->opts.rewindBufferCapacity > 0) {
-		if (thread->state != THREAD_REWINDING) {
-			mCoreRewindAppend(&thread->rewind, thread->core);
-		} else if (thread->state == THREAD_REWINDING) {
-			if (!mCoreRewindRestore(&thread->rewind, thread->core)) {
-				mCoreRewindAppend(&thread->rewind, thread->core);
-			}
-		}
-	}
-}
-
-void mCoreThreadFrameEnded(struct mCoreThread* thread) {
-	if (!thread) {
-		return;
-	}
-	if (thread->frameCallback) {
-		thread->frameCallback(thread);
-	}
-}
-
 #else
 struct mCoreThread* mCoreThreadGet(void) {
 	return NULL;
 }
-
-void mCoreThreadFrameStarted(struct mCoreThread* thread) {
-	UNUSED(thread);
-}
-
-void mCoreThreadFrameEnded(struct mCoreThread* thread) {
-	UNUSED(thread);
-}
-
 #endif
 
 static void _mCoreLog(struct mLogger* logger, int category, enum mLogLevel level, const char* format, va_list args) {
 	UNUSED(logger);
-	struct mCoreThread* thread = mCoreThreadGet();
-	if (thread && !(thread->logLevel & level)) {
-		return;
-	}
+	UNUSED(level);
 	printf("%s: ", mLogCategoryName(category));
 	vprintf(format, args);
 	printf("\n");
@@ -536,9 +581,6 @@ static void _mCoreLog(struct mLogger* logger, int category, enum mLogLevel level
 struct mLogger* mCoreThreadLogger(void) {
 	struct mCoreThread* thread = mCoreThreadGet();
 	if (thread) {
-		if (!thread->logger.d.log) {
-			thread->logger.d.log = _mCoreLog;
-		}
 		return &thread->logger.d;
 	}
 	return NULL;
