@@ -3,12 +3,12 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "gdb-stub.h"
+#include <mgba/internal/debugger/gdb-stub.h>
 
-#include "arm/debugger/debugger.h"
-#include "arm/isa-inlines.h"
-#include "core/core.h"
-#include "gba/memory.h"
+#include <mgba/core/core.h>
+#include <mgba/internal/arm/debugger/debugger.h>
+#include <mgba/internal/arm/isa-inlines.h>
+#include <mgba/internal/gba/memory.h>
 
 #include <signal.h>
 
@@ -46,7 +46,7 @@ static void _gdbStubEntered(struct mDebugger* debugger, enum mDebuggerEntryReaso
 		break;
 	case DEBUGGER_ENTER_BREAKPOINT:
 		if (stub->supportsHwbreak && stub->supportsSwbreak && info) {
-			snprintf(stub->outgoing, GDB_STUB_MAX_LINE - 4, "T%02x%cwbreak:;", SIGTRAP, info->breakType == BREAKPOINT_SOFTWARE ? 's' : 'h');
+			snprintf(stub->outgoing, GDB_STUB_MAX_LINE - 4, "T%02x%cwbreak:;", SIGTRAP, info->type.bp.breakType == BREAKPOINT_SOFTWARE ? 's' : 'h');
 		} else {
 			snprintf(stub->outgoing, GDB_STUB_MAX_LINE - 4, "S%02xk", SIGTRAP);
 		}
@@ -54,9 +54,9 @@ static void _gdbStubEntered(struct mDebugger* debugger, enum mDebuggerEntryReaso
 	case DEBUGGER_ENTER_WATCHPOINT:
 		if (info) {
 			const char* type = 0;
-			switch (info->watchType) {
+			switch (info->type.wp.watchType) {
 			case WATCHPOINT_WRITE:
-				if (info->newValue == info->oldValue) {
+				if (info->type.wp.newValue == info->type.wp.oldValue) {
 					if (stub->d.state == DEBUGGER_PAUSED) {
 						stub->d.state = DEBUGGER_RUNNING;
 					}
@@ -321,13 +321,31 @@ static void _readGPRs(struct GDBStub* stub, const char* message) {
 	UNUSED(message);
 	int r;
 	int i = 0;
+
+	// General purpose registers
 	for (r = 0; r < ARM_PC; ++r) {
 		_int2hex32(cpu->gprs[r], &stub->outgoing[i]);
 		i += 8;
 	}
+
+	// Program counter
 	_int2hex32(cpu->gprs[ARM_PC] - (cpu->cpsr.t ? WORD_SIZE_THUMB : WORD_SIZE_ARM), &stub->outgoing[i]);
 	i += 8;
 
+	// Floating point registers, unused on the GBA (8 of them, 24 bits each)
+	for (r = 0; r < 8 * 3; ++r) {
+		_int2hex32(0, &stub->outgoing[i]);
+		i += 8;
+	}
+
+	// Floating point status, unused on the GBA (32 bits)
+	_int2hex32(0, &stub->outgoing[i]);
+	i += 8;
+
+	// CPU status
+	_int2hex32(cpu->cpsr.packed, &stub->outgoing[i]);
+	i += 8;
+
 	stub->outgoing[i] = 0;
 	_sendMessage(stub);
 }
@@ -345,7 +363,7 @@ static void _writeRegister(struct GDBStub* stub, const char* message) {
 #ifdef _MSC_VER
 	value = _byteswap_ulong(value);
 #else
-	value = __builtin_bswap32(value);
+	LOAD_32BE(value, 0, &value);
 #endif
 
 	if (reg <= ARM_PC) {
@@ -477,16 +495,16 @@ static void _setBreakpoint(struct GDBStub* stub, const char* message) {
 		ARMDebuggerSetSoftwareBreakpoint(stub->d.platform, address, kind == 2 ? MODE_THUMB : MODE_ARM);
 		break;
 	case '1':
-		stub->d.platform->setBreakpoint(stub->d.platform, address);
+		stub->d.platform->setBreakpoint(stub->d.platform, address, -1);
 		break;
 	case '2':
-		stub->d.platform->setWatchpoint(stub->d.platform, address, WATCHPOINT_WRITE);
+		stub->d.platform->setWatchpoint(stub->d.platform, address, -1, WATCHPOINT_WRITE);
 		break;
 	case '3':
-		stub->d.platform->setWatchpoint(stub->d.platform, address, WATCHPOINT_READ);
+		stub->d.platform->setWatchpoint(stub->d.platform, address, -1, WATCHPOINT_READ);
 		break;
 	case '4':
-		stub->d.platform->setWatchpoint(stub->d.platform, address, WATCHPOINT_RW);
+		stub->d.platform->setWatchpoint(stub->d.platform, address, -1, WATCHPOINT_RW);
 		break;
 	default:
 		stub->outgoing[0] = '\0';
@@ -506,12 +524,12 @@ static void _clearBreakpoint(struct GDBStub* stub, const char* message) {
 		ARMDebuggerClearSoftwareBreakpoint(stub->d.platform, address);
 		break;
 	case '1':
-		stub->d.platform->clearBreakpoint(stub->d.platform, address);
+		stub->d.platform->clearBreakpoint(stub->d.platform, address, -1);
 		break;
 	case '2':
 	case '3':
 	case '4':
-		stub->d.platform->clearWatchpoint(stub->d.platform, address);
+		stub->d.platform->clearWatchpoint(stub->d.platform, address, -1);
 		break;
 	default:
 		break;
