@@ -5,8 +5,8 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #include "ffmpeg-encoder.h"
 
-#include "core/core.h"
-#include "gba/video.h"
+#include <mgba/core/core.h>
+#include <mgba/internal/gba/video.h>
 
 #include <libavcodec/version.h>
 #include <libavcodec/avcodec.h>
@@ -229,7 +229,11 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		AVDictionary* opts = 0;
 		av_dict_set(&opts, "strict", "-2", 0);
 		if (encoder->context->oformat->flags & AVFMT_GLOBALHEADER) {
+#ifdef AV_CODEC_FLAG_GLOBAL_HEADER
+			encoder->audio->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
+#else
 			encoder->audio->flags |= CODEC_FLAG_GLOBAL_HEADER;
+#endif
 		}
 		avcodec_open2(encoder->audio, acodec, &opts);
 		av_dict_free(&opts);
@@ -291,7 +295,11 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 	encoder->video->gop_size = 60;
 	encoder->video->max_b_frames = 3;
 	if (encoder->context->oformat->flags & AVFMT_GLOBALHEADER) {
+#ifdef AV_CODEC_FLAG_GLOBAL_HEADER
+		encoder->video->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
+#else
 		encoder->video->flags |= CODEC_FLAG_GLOBAL_HEADER;
+#endif
 	}
 	if (strcmp(vcodec->name, "libx264") == 0) {
 		// Try to adaptively figure out when you can use a slower encoder
@@ -304,6 +312,14 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		}
 		av_opt_set(encoder->video->priv_data, "tune", "zerolatency", 0);
 	}
+
+	if (encoder->video->codec->id == AV_CODEC_ID_H264 &&
+	    (strcasecmp(encoder->containerFormat, "mp4") ||
+	        strcasecmp(encoder->containerFormat, "m4v") ||
+	        strcasecmp(encoder->containerFormat, "mov"))) {
+		// QuickTime and a few other things require YUV420
+		encoder->video->pix_fmt = AV_PIX_FMT_YUV420P;
+	}
 	avcodec_open2(encoder->video, vcodec, 0);
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 	encoder->videoFrame = av_frame_alloc();
@@ -320,7 +336,9 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 	avcodec_parameters_from_context(encoder->videoStream->codecpar, encoder->video);
 #endif
 
-	avio_open(&encoder->context->pb, outfile, AVIO_FLAG_WRITE);
+	if (avio_open(&encoder->context->pb, outfile, AVIO_FLAG_WRITE) < 0) {
+		return false;
+	}
 	return avformat_write_header(encoder->context, 0) >= 0;
 }
 
@@ -389,27 +407,27 @@ void _ffmpegPostAudioFrame(struct mAVStream* stream, int16_t left, int16_t right
 	encoder->audioBuffer[encoder->currentAudioSample * 2] = left;
 	encoder->audioBuffer[encoder->currentAudioSample * 2 + 1] = right;
 
-	++encoder->currentAudioFrame;
 	++encoder->currentAudioSample;
 
-	if ((encoder->currentAudioSample * 4) < encoder->audioBufferSize) {
+	if (encoder->currentAudioSample * 4 < encoder->audioBufferSize) {
 		return;
 	}
 
 	int channelSize = 2 * av_get_bytes_per_sample(encoder->audio->sample_fmt);
-	avresample_convert(encoder->resampleContext,
-	    0, 0, 0,
-	    (uint8_t**) &encoder->audioBuffer, 0, encoder->audioBufferSize / 4);
+	avresample_convert(encoder->resampleContext, 0, 0, 0,
+	                   (uint8_t**) &encoder->audioBuffer, 0, encoder->audioBufferSize / 4);
+
+	encoder->currentAudioSample = 0;
 	if (avresample_available(encoder->resampleContext) < encoder->audioFrame->nb_samples) {
 		return;
 	}
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 	av_frame_make_writable(encoder->audioFrame);
 #endif
-	avresample_read(encoder->resampleContext, encoder->audioFrame->data, encoder->postaudioBufferSize / channelSize);
+	int samples = avresample_read(encoder->resampleContext, encoder->audioFrame->data, encoder->postaudioBufferSize / channelSize);
 
-	encoder->audioFrame->pts = av_rescale_q(encoder->currentAudioFrame - encoder->currentAudioSample, encoder->audio->time_base, encoder->audioStream->time_base);
-	encoder->currentAudioSample = 0;
+	encoder->audioFrame->pts = av_rescale_q(encoder->currentAudioFrame, encoder->audio->time_base, encoder->audioStream->time_base);
+	encoder->currentAudioFrame += samples;
 
 	AVPacket packet;
 	av_init_packet(&packet);
