@@ -5,18 +5,18 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #include "gui-runner.h"
 
-#include "core/core.h"
-#include "core/serialize.h"
+#include <mgba/core/core.h>
+#include <mgba/core/serialize.h>
 #include "feature/gui/gui-config.h"
-#include "gba/gba.h"
-#include "gba/input.h"
-#include "gba/interface.h"
-#include "util/gui/file-select.h"
-#include "util/gui/font.h"
-#include "util/gui/menu.h"
-#include "util/memory.h"
-#include "util/png-io.h"
-#include "util/vfs.h"
+#include <mgba/internal/gba/gba.h>
+#include <mgba/internal/gba/input.h>
+#include <mgba/gba/interface.h>
+#include <mgba-util/gui/file-select.h>
+#include <mgba-util/gui/font.h>
+#include <mgba-util/gui/menu.h>
+#include <mgba-util/memory.h>
+#include <mgba-util/png-io.h>
+#include <mgba-util/vfs.h>
 
 #ifdef _3DS
 #include <3ds.h>
@@ -25,7 +25,7 @@
 #include <sys/time.h>
 
 mLOG_DECLARE_CATEGORY(GUI_RUNNER);
-mLOG_DEFINE_CATEGORY(GUI_RUNNER, "GUI Runner");
+mLOG_DEFINE_CATEGORY(GUI_RUNNER, "GUI Runner", "gui.runner");
 
 #define FPS_GRANULARITY 120
 #define FPS_BUFFER_SIZE 3
@@ -165,12 +165,19 @@ void mGUIInit(struct mGUIRunner* runner, const char* port) {
 	mCoreConfigSetDefaultIntValue(&runner->config, "volume", 0x100);
 	mCoreConfigSetDefaultValue(&runner->config, "idleOptimization", "detect");
 	mCoreConfigLoad(&runner->config);
+	mCoreConfigGetIntValue(&runner->config, "logLevel", &logger.logLevel);
 
 	char path[PATH_MAX];
 	mCoreConfigDirectory(path, PATH_MAX);
 	strncat(path, PATH_SEP "log", PATH_MAX - strlen(path));
 	logger.vf = VFileOpen(path, O_CREAT | O_WRONLY | O_APPEND);
 	mLogSetDefaultLogger(&logger.d);
+
+	const char* lastPath = mCoreConfigGetValue(&runner->config, "lastDirectory");
+	if (lastPath) {
+		strncpy(runner->params.currentPath, lastPath, PATH_MAX - 1);
+		runner->params.currentPath[PATH_MAX - 1] = '\0';
+	}
 }
 
 void mGUIDeinit(struct mGUIRunner* runner) {
@@ -277,6 +284,7 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 	if (runner->core) {
 		mLOG(GUI_RUNNER, INFO, "Found core");
 		runner->core->init(runner->core);
+		mCoreInitConfig(runner->core, runner->port);
 		mInputMapInit(&runner->core->inputMap, &GBAInputInfo);
 		found = mCoreLoadFile(runner->core, path);
 		if (!found) {
@@ -291,11 +299,10 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 		return;
 	}
 	if (runner->core->platform(runner->core) == PLATFORM_GBA) {
-		((struct GBA*) runner->core->board)->luminanceSource = &runner->luminanceSource.d;
+		runner->core->setPeripheral(runner->core, mPERIPH_GBA_LUMINANCE, &runner->luminanceSource.d);
 	}
 	mLOG(GUI_RUNNER, DEBUG, "Loading config...");
 	mCoreLoadForeignConfig(runner->core, &runner->config);
-	logger.logLevel = runner->core->opts.logLevel;
 
 	mLOG(GUI_RUNNER, DEBUG, "Loading save...");
 	mCoreAutoloadSave(runner->core);
@@ -424,10 +431,10 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 				runner->core->reset(runner->core);
 				break;
 			case RUNNER_SAVE_STATE:
-				mCoreSaveState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_SAVEDATA);
+				mCoreSaveState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_SAVEDATA | SAVESTATE_RTC | SAVESTATE_METADATA);
 				break;
 			case RUNNER_LOAD_STATE:
-				mCoreLoadState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT);
+				mCoreLoadState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_RTC);
 				break;
 			case RUNNER_SCREENSHOT:
 				mCoreTakeScreenshot(runner->core);
@@ -499,6 +506,8 @@ void mGUIRunloop(struct mGUIRunner* runner) {
 		if (!GUISelectFile(&runner->params, path, sizeof(path), 0)) {
 			break;
 		}
+		mCoreConfigSetValue(&runner->config, "lastDirectory", runner->params.currentPath);
+		mCoreConfigSave(&runner->config);
 		mGUIRun(runner, path);
 	}
 }
