@@ -3,18 +3,19 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "gb.h"
+#include <mgba/internal/gb/gb.h>
 
-#include "gb/io.h"
-#include "gb/mbc.h"
+#include <mgba/internal/gb/io.h>
+#include <mgba/internal/gb/mbc.h>
+#include <mgba/internal/lr35902/lr35902.h>
 
-#include "core/core.h"
-#include "core/cheats.h"
-#include "util/crc32.h"
-#include "util/memory.h"
-#include "util/math.h"
-#include "util/patch.h"
-#include "util/vfs.h"
+#include <mgba/core/core.h>
+#include <mgba/core/cheats.h>
+#include <mgba-util/crc32.h>
+#include <mgba-util/memory.h>
+#include <mgba-util/math.h>
+#include <mgba-util/patch.h>
+#include <mgba-util/vfs.h>
 
 #define CLEANUP_THRESHOLD 15
 
@@ -23,19 +24,24 @@ const uint32_t SGB_LR35902_FREQUENCY = 0x418B1E;
 
 const uint32_t GB_COMPONENT_MAGIC = 0x400000;
 
+static const uint8_t _knownHeader[4] = { 0xCE, 0xED, 0x66, 0x66};
+
 #define DMG_BIOS_CHECKSUM 0xC2F5CC97
 #define DMG_2_BIOS_CHECKSUM 0x59C8598E
 #define CGB_BIOS_CHECKSUM 0x41884E46
 
-mLOG_DEFINE_CATEGORY(GB, "GB");
+mLOG_DEFINE_CATEGORY(GB, "GB", "gb");
 
 static void GBInit(void* cpu, struct mCPUComponent* component);
+static void GBDeinit(struct mCPUComponent* component);
 static void GBInterruptHandlerInit(struct LR35902InterruptHandler* irqh);
 static void GBProcessEvents(struct LR35902Core* cpu);
 static void GBSetInterrupts(struct LR35902Core* cpu, bool enable);
 static void GBIllegal(struct LR35902Core* cpu);
 static void GBStop(struct LR35902Core* cpu);
 
+static void _enableInterrupts(struct mTiming* timing, void* user, uint32_t cyclesLate);
+
 #ifdef _3DS
 extern uint32_t* romBuffer;
 extern size_t romBufferSize;
@@ -44,7 +50,7 @@ extern size_t romBufferSize;
 void GBCreate(struct GB* gb) {
 	gb->d.id = GB_COMPONENT_MAGIC;
 	gb->d.init = GBInit;
-	gb->d.deinit = 0;
+	gb->d.deinit = GBDeinit;
 }
 
 static void GBInit(void* cpu, struct mCPUComponent* component) {
@@ -73,11 +79,25 @@ static void GBInit(void* cpu, struct mCPUComponent* component) {
 	gb->sramVf = NULL;
 	gb->sramRealVf = NULL;
 
-	gb->pristineRom = 0;
+	gb->isPristine = false;
 	gb->pristineRomSize = 0;
 	gb->yankedRomSize = 0;
 
+	mCoreCallbacksListInit(&gb->coreCallbacks, 0);
 	gb->stream = NULL;
+
+	mTimingInit(&gb->timing, &gb->cpu->cycles, &gb->cpu->nextEvent);
+	gb->audio.timing = &gb->timing;
+
+	gb->eiPending.name = "GB EI";
+	gb->eiPending.callback = _enableInterrupts;
+	gb->eiPending.context = gb;
+	gb->eiPending.priority = 0;
+}
+
+static void GBDeinit(struct mCPUComponent* component) {
+	struct GB* gb = (struct GB*) component;
+	mTimingDeinit(&gb->timing);
 }
 
 bool GBLoadROM(struct GB* gb, struct VFile* vf) {
@@ -88,24 +108,23 @@ bool GBLoadROM(struct GB* gb, struct VFile* vf) {
 	gb->romVf = vf;
 	gb->pristineRomSize = vf->size(vf);
 	vf->seek(vf, 0, SEEK_SET);
+	gb->isPristine = true;
 #ifdef _3DS
-	gb->pristineRom = 0;
 	if (gb->pristineRomSize <= romBufferSize) {
-		gb->pristineRom = romBuffer;
+		gb->memory.rom = romBuffer;
 		vf->read(vf, romBuffer, gb->pristineRomSize);
 	}
 #else
-	gb->pristineRom = vf->map(vf, gb->pristineRomSize, MAP_READ);
+	gb->memory.rom = vf->map(vf, gb->pristineRomSize, MAP_READ);
 #endif
-	if (!gb->pristineRom) {
+	if (!gb->memory.rom) {
 		return false;
 	}
 	gb->yankedRomSize = 0;
-	gb->memory.rom = gb->pristineRom;
 	gb->memory.romBase = gb->memory.rom;
 	gb->memory.romSize = gb->pristineRomSize;
 	gb->romCrc32 = doCrc32(gb->memory.rom, gb->memory.romSize);
-	GBMBCSwitchBank(&gb->memory, gb->memory.currentBank);
+	GBMBCInit(gb);
 
 	if (gb->cpu) {
 		struct LR35902Core* cpu = gb->cpu;
@@ -116,12 +135,6 @@ bool GBLoadROM(struct GB* gb, struct VFile* vf) {
 	return true;
 }
 
-bool GBLoadSave(struct GB* gb, struct VFile* vf) {
-	gb->sramVf = vf;
-	gb->sramRealVf = vf;
-	return vf;
-}
-
 static void GBSramDeinit(struct GB* gb) {
 	if (gb->sramVf) {
 		gb->sramVf->unmap(gb->sramVf, gb->memory.sram, gb->sramSize);
@@ -135,11 +148,20 @@ static void GBSramDeinit(struct GB* gb) {
 	gb->memory.sram = 0;
 }
 
+bool GBLoadSave(struct GB* gb, struct VFile* vf) {
+	GBSramDeinit(gb);
+	gb->sramVf = vf;
+	gb->sramRealVf = vf;
+	if (gb->sramSize) {
+		GBResizeSram(gb, gb->sramSize);
+	}
+	return vf;
+}
+
 void GBResizeSram(struct GB* gb, size_t size) {
 	if (gb->memory.sram && size <= gb->sramSize) {
 		return;
 	}
-	mLOG(GB, INFO, "Resizing SRAM to %"PRIz"u bytes", size);
 	struct VFile* vf = gb->sramVf;
 	if (vf) {
 		if (vf == gb->sramRealVf) {
@@ -197,7 +219,7 @@ void GBResizeSram(struct GB* gb, size_t size) {
 
 void GBSramClean(struct GB* gb, uint32_t frameCount) {
 	// TODO: Share with GBASavedataClean
-	if (!gb->sramVf || gb->sramVf != gb->sramRealVf) {
+	if (!gb->sramVf) {
 		return;
 	}
 	if (gb->sramDirty & GB_SRAM_DIRT_NEW) {
@@ -207,6 +229,9 @@ void GBSramClean(struct GB* gb, uint32_t frameCount) {
 			gb->sramDirty |= GB_SRAM_DIRT_SEEN;
 		}
 	} else if ((gb->sramDirty & GB_SRAM_DIRT_SEEN) && frameCount - gb->sramDirtAge > CLEANUP_THRESHOLD) {
+		if (gb->sramMaskWriteback) {
+			GBSavedataUnmask(gb);
+		}
 		if (gb->memory.mbcType == GB_MBC3_RTC) {
 			GBMBCRTCWrite(gb);
 		}
@@ -236,34 +261,37 @@ void GBSavedataUnmask(struct GB* gb) {
 	gb->sramVf = gb->sramRealVf;
 	gb->memory.sram = gb->sramVf->map(gb->sramVf, gb->sramSize, MAP_WRITE);
 	if (gb->sramMaskWriteback) {
+		vf->seek(vf, 0, SEEK_SET);
 		vf->read(vf, gb->memory.sram, gb->sramSize);
+		gb->sramMaskWriteback = false;
 	}
 	vf->close(vf);
 }
 
 void GBUnloadROM(struct GB* gb) {
 	// TODO: Share with GBAUnloadROM
-	if (gb->memory.rom && gb->memory.romBase != gb->memory.rom && gb->memory.romBase != gb->pristineRom) {
+	if (gb->memory.rom && gb->memory.romBase != gb->memory.rom && !gb->isPristine) {
 		free(gb->memory.romBase);
 	}
-	if (gb->memory.rom && gb->pristineRom != gb->memory.rom) {
+	if (gb->memory.rom && !gb->isPristine) {
 		if (gb->yankedRomSize) {
 			gb->yankedRomSize = 0;
 		}
 		mappedMemoryFree(gb->memory.rom, GB_SIZE_CART_MAX);
-		gb->memory.rom = gb->pristineRom;
 	}
-	gb->memory.rom = 0;
 
 	if (gb->romVf) {
 #ifndef _3DS
-		gb->romVf->unmap(gb->romVf, gb->pristineRom, gb->pristineRomSize);
+		gb->romVf->unmap(gb->romVf, gb->memory.rom, gb->pristineRomSize);
 #endif
 		gb->romVf->close(gb->romVf);
-		gb->romVf = 0;
+		gb->romVf = NULL;
 	}
-	gb->pristineRom = 0;
+	gb->memory.rom = NULL;
+	gb->memory.mbcType = GB_MBC_AUTODETECT;
+	gb->isPristine = false;
 
+	gb->sramMaskWriteback = false;
 	GBSavedataUnmask(gb);
 	GBSramDeinit(gb);
 	if (gb->sramRealVf) {
@@ -273,6 +301,18 @@ void GBUnloadROM(struct GB* gb) {
 	gb->sramVf = NULL;
 }
 
+void GBSynthesizeROM(struct VFile* vf) {
+	if (!vf) {
+		return;
+	}
+	const struct GBCartridge cart = {
+		.logo = { _knownHeader[0], _knownHeader[1], _knownHeader[2], _knownHeader[3]}
+	};
+
+	vf->seek(vf, 0x100, SEEK_SET);
+	vf->write(vf, &cart, sizeof(cart));
+}
+
 void GBLoadBIOS(struct GB* gb, struct VFile* vf) {
 	gb->biosVf = vf;
 }
@@ -285,14 +325,26 @@ void GBApplyPatch(struct GB* gb, struct Patch* patch) {
 	if (patchedSize > GB_SIZE_CART_MAX) {
 		patchedSize = GB_SIZE_CART_MAX;
 	}
-	gb->memory.rom = anonymousMemoryMap(GB_SIZE_CART_MAX);
-	if (!patch->applyPatch(patch, gb->pristineRom, gb->pristineRomSize, gb->memory.rom, patchedSize)) {
-		mappedMemoryFree(gb->memory.rom, patchedSize);
-		gb->memory.rom = gb->pristineRom;
+	void* newRom = anonymousMemoryMap(GB_SIZE_CART_MAX);
+	if (!patch->applyPatch(patch, gb->memory.rom, gb->pristineRomSize, newRom, patchedSize)) {
+		mappedMemoryFree(newRom, GB_SIZE_CART_MAX);
 		return;
 	}
+	if (gb->romVf) {
+#ifndef _3DS
+		gb->romVf->unmap(gb->romVf, gb->memory.rom, gb->pristineRomSize);
+#endif
+		gb->romVf->close(gb->romVf);
+		gb->romVf = NULL;
+	}
+	gb->isPristine = false;
+	if (gb->memory.romBase == gb->memory.rom) {
+		gb->memory.romBase = newRom;
+	}
+	gb->memory.rom = newRom;
 	gb->memory.romSize = patchedSize;
 	gb->romCrc32 = doCrc32(gb->memory.rom, gb->memory.romSize);
+	gb->cpu->memory.setActiveRegion(gb->cpu, gb->cpu->pc);
 }
 
 void GBDestroy(struct GB* gb) {
@@ -307,6 +359,7 @@ void GBDestroy(struct GB* gb) {
 	GBAudioDeinit(&gb->audio);
 	GBVideoDeinit(&gb->video);
 	GBSIODeinit(&gb->sio);
+	mCoreCallbacksListDeinit(&gb->coreCallbacks);
 }
 
 void GBInterruptHandlerInit(struct LR35902InterruptHandler* irqh) {
@@ -342,7 +395,9 @@ bool GBIsBIOS(struct VFile* vf) {
 
 void GBReset(struct LR35902Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
+	gb->memory.romBase = gb->memory.rom;
 	GBDetectModel(gb);
+
 	if (gb->biosVf) {
 		if (!GBIsBIOS(gb->biosVf)) {
 			gb->biosVf->close(gb->biosVf);
@@ -383,6 +438,7 @@ void GBReset(struct LR35902Core* cpu) {
 			cpu->e = 0xD8;
 			cpu->h = 1;
 			cpu->l = 0x4D;
+			gb->timer.internalDiv = 0x2AF3;
 			break;
 		case GB_MODEL_AGB:
 			cpu->b = 1;
@@ -394,6 +450,7 @@ void GBReset(struct LR35902Core* cpu) {
 			cpu->e = 0x08;
 			cpu->h = 0;
 			cpu->l = 0x7C;
+			gb->timer.internalDiv = 0x7A8;
 			break;
 		}
 
@@ -401,7 +458,8 @@ void GBReset(struct LR35902Core* cpu) {
 		cpu->pc = 0x100;
 	}
 
-	gb->eiPending = INT_MAX;
+	gb->cpuBlocked = false;
+	gb->earlyExit = false;
 	gb->doubleSpeed = 0;
 
 	cpu->memory.setActiveRegion(cpu, cpu->pc);
@@ -410,9 +468,14 @@ void GBReset(struct LR35902Core* cpu) {
 		gb->memory.romSize = gb->yankedRomSize;
 		gb->yankedRomSize = 0;
 	}
+
+	mTimingClear(&gb->timing);
+
 	GBMemoryReset(gb);
 	GBVideoReset(&gb->video);
 	GBTimerReset(&gb->timer);
+	mTimingSchedule(&gb->timing, &gb->timer.event, GB_DMG_DIV_PERIOD);
+
 	GBAudioReset(&gb->audio);
 	GBIOReset(gb);
 	GBSIOReset(&gb->sio);
@@ -500,55 +563,22 @@ void GBUpdateIRQs(struct GB* gb) {
 void GBProcessEvents(struct LR35902Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	do {
-		int32_t cycles = cpu->nextEvent;
-		int32_t nextEvent = INT_MAX;
-		int32_t testEvent;
+		int32_t cycles = cpu->cycles;
+		int32_t nextEvent;
 
-		if (gb->eiPending != INT_MAX) {
-			gb->eiPending -= cycles;
-			if (gb->eiPending <= 0) {
-				gb->memory.ime = true;
-				GBUpdateIRQs(gb);
-				gb->eiPending = INT_MAX;
-			} else {
-				nextEvent = gb->eiPending;
-			}
-		}
+		cpu->cycles = 0;
+		cpu->nextEvent = INT_MAX;
 
-		testEvent = GBVideoProcessEvents(&gb->video, cycles >> gb->doubleSpeed);
-		if (testEvent != INT_MAX) {
-			testEvent <<= gb->doubleSpeed;
-			if (testEvent < nextEvent) {
-				nextEvent = testEvent;
-			}
-		}
-
-		testEvent = GBAudioProcessEvents(&gb->audio, cycles >> gb->doubleSpeed);
-		if (testEvent != INT_MAX) {
-			testEvent <<= gb->doubleSpeed;
-			if (testEvent < nextEvent) {
-				nextEvent = testEvent;
-			}
-		}
-
-		testEvent = GBTimerProcessEvents(&gb->timer, cycles);
-		if (testEvent < nextEvent) {
-			nextEvent = testEvent;
-		}
-
-		testEvent = GBSIOProcessEvents(&gb->sio, cycles);
-		if (testEvent < nextEvent) {
-			nextEvent = testEvent;
-		}
-
-		testEvent = GBMemoryProcessEvents(gb, cycles);
-		if (testEvent < nextEvent) {
-			nextEvent = testEvent;
-		}
-
-		cpu->cycles -= cycles;
+		nextEvent = cycles;
+		do {
+			nextEvent = mTimingTick(&gb->timing, nextEvent);
+		} while (gb->cpuBlocked);
 		cpu->nextEvent = nextEvent;
 
+		if (gb->earlyExit) {
+			gb->earlyExit = false;
+			break;
+		}
 		if (cpu->halted) {
 			cpu->cycles = cpu->nextEvent;
 			if (!gb->memory.ie || !gb->memory.ime) {
@@ -562,16 +592,22 @@ void GBSetInterrupts(struct LR35902Core* cpu, bool enable) {
 	struct GB* gb = (struct GB*) cpu->master;
 	if (!enable) {
 		gb->memory.ime = enable;
-		gb->eiPending = INT_MAX;
+		mTimingDeschedule(&gb->timing, &gb->eiPending);
 		GBUpdateIRQs(gb);
 	} else {
-		if (cpu->nextEvent > cpu->cycles + 4) {
-			cpu->nextEvent = cpu->cycles + 4;
-		}
-		gb->eiPending = cpu->cycles + 4;
+		mTimingDeschedule(&gb->timing, &gb->eiPending);
+		mTimingSchedule(&gb->timing, &gb->eiPending, 4);
 	}
 }
 
+static void _enableInterrupts(struct mTiming* timing, void* user, uint32_t cyclesLate) {
+	UNUSED(timing);
+	UNUSED(cyclesLate);
+	struct GB* gb = user;
+	gb->memory.ime = true;
+	GBUpdateIRQs(gb);
+}
+
 void GBHalt(struct LR35902Core* cpu) {
 	if (!cpu->irqPending) {
 		cpu->cycles = cpu->nextEvent;
@@ -586,16 +622,19 @@ void GBStop(struct LR35902Core* cpu) {
 	}
 	if (gb->memory.io[REG_KEY1] & 1) {
 		gb->doubleSpeed ^= 1;
+		gb->audio.timingFactor = gb->doubleSpeed + 1;
 		gb->memory.io[REG_KEY1] = 0;
 		gb->memory.io[REG_KEY1] |= gb->doubleSpeed << 7;
 	} else if (cpu->bus) {
+#ifdef USE_DEBUGGERS
 		if (cpu->components && cpu->components[CPU_COMPONENT_DEBUGGER]) {
 			struct mDebuggerEntryInfo info = {
 				.address = cpu->pc - 1,
-				.opcode = 0x1000 | cpu->bus
+				.type.bp.opcode = 0x1000 | cpu->bus
 			};
 			mDebuggerEnter((struct mDebugger*) cpu->components[CPU_COMPONENT_DEBUGGER], DEBUGGER_ENTER_ILLEGAL_OP, &info);
 		}
+#endif
 		// Hang forever
 		gb->memory.ime = 0;
 		cpu->pc -= 2;
@@ -606,40 +645,41 @@ void GBStop(struct LR35902Core* cpu) {
 void GBIllegal(struct LR35902Core* cpu) {
 	struct GB* gb = (struct GB*) cpu->master;
 	mLOG(GB, GAME_ERROR, "Hit illegal opcode at address %04X:%02X\n", cpu->pc, cpu->bus);
+#ifdef USE_DEBUGGERS
 	if (cpu->components && cpu->components[CPU_COMPONENT_DEBUGGER]) {
 		struct mDebuggerEntryInfo info = {
 			.address = cpu->pc,
-			.opcode = cpu->bus
+			.type.bp.opcode = cpu->bus
 		};
 		mDebuggerEnter((struct mDebugger*) cpu->components[CPU_COMPONENT_DEBUGGER], DEBUGGER_ENTER_ILLEGAL_OP, &info);
 	}
+#endif
 	// Hang forever
 	gb->memory.ime = 0;
 	--cpu->pc;
 }
 
 bool GBIsROM(struct VFile* vf) {
+	if (!vf) {
+		return false;
+	}
 	vf->seek(vf, 0x104, SEEK_SET);
 	uint8_t header[4];
-	static const uint8_t knownHeader[4] = { 0xCE, 0xED, 0x66, 0x66};
 
 	if (vf->read(vf, &header, sizeof(header)) < (ssize_t) sizeof(header)) {
 		return false;
 	}
-	if (memcmp(header, knownHeader, sizeof(header))) {
+	if (memcmp(header, _knownHeader, sizeof(header))) {
 		return false;
 	}
 	return true;
 }
 
-void GBGetGameTitle(struct GB* gb, char* out) {
+void GBGetGameTitle(const struct GB* gb, char* out) {
 	const struct GBCartridge* cart = NULL;
 	if (gb->memory.rom) {
 		cart = (const struct GBCartridge*) &gb->memory.rom[0x100];
 	}
-	if (gb->pristineRom) {
-		cart = (const struct GBCartridge*) &((uint8_t*) gb->pristineRom)[0x100];
-	}
 	if (!cart) {
 		return;
 	}
@@ -650,15 +690,12 @@ void GBGetGameTitle(struct GB* gb, char* out) {
 	}
 }
 
-void GBGetGameCode(struct GB* gb, char* out) {
+void GBGetGameCode(const struct GB* gb, char* out) {
 	memset(out, 0, 8);
 	const struct GBCartridge* cart = NULL;
 	if (gb->memory.rom) {
 		cart = (const struct GBCartridge*) &gb->memory.rom[0x100];
 	}
-	if (gb->pristineRom) {
-		cart = (const struct GBCartridge*) &((uint8_t*) gb->pristineRom)[0x100];
-	}
 	if (!cart) {
 		return;
 	}
@@ -683,4 +720,6 @@ void GBFrameEnded(struct GB* gb) {
 			mCheatRefresh(device, cheats);
 		}
 	}
+
+	GBTestKeypadIRQ(gb);
 }
