@@ -3,13 +3,15 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "mbc.h"
+#include <mgba/internal/gb/mbc.h>
 
-#include "gb/gb.h"
-#include "gb/memory.h"
-#include "util/vfs.h"
+#include <mgba/core/interface.h>
+#include <mgba/internal/lr35902/lr35902.h>
+#include <mgba/internal/gb/gb.h>
+#include <mgba/internal/gb/memory.h>
+#include <mgba-util/vfs.h>
 
-mLOG_DEFINE_CATEGORY(GB_MBC, "GB MBC");
+mLOG_DEFINE_CATEGORY(GB_MBC, "GB MBC", "gb.mbc");
 
 static void _GBMBCNone(struct GB* gb, uint16_t address, uint8_t value) {
 	UNUSED(gb);
@@ -26,21 +28,60 @@ static void _GBMBC5(struct GB*, uint16_t address, uint8_t value);
 static void _GBMBC6(struct GB*, uint16_t address, uint8_t value);
 static void _GBMBC7(struct GB*, uint16_t address, uint8_t value);
 static void _GBHuC3(struct GB*, uint16_t address, uint8_t value);
+static void _GBPocketCam(struct GB* gb, uint16_t address, uint8_t value);
 
-void GBMBCSwitchBank(struct GBMemory* memory, int bank) {
+static uint8_t _GBMBC2Read(struct GBMemory*, uint16_t address);
+static uint8_t _GBMBC7Read(struct GBMemory*, uint16_t address);
+static uint8_t _GBPocketCamRead(struct GBMemory*, uint16_t address);
+
+void GBMBCSwitchBank(struct GB* gb, int bank) {
 	size_t bankStart = bank * GB_SIZE_CART_BANK0;
-	if (bankStart + GB_SIZE_CART_BANK0 > memory->romSize) {
+	if (bankStart + GB_SIZE_CART_BANK0 > gb->memory.romSize) {
 		mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid ROM bank: %0X", bank);
-		bankStart &= (memory->romSize - 1);
+		bankStart &= (gb->memory.romSize - 1);
 		bank = bankStart / GB_SIZE_CART_BANK0;
 	}
-	memory->romBank = &memory->rom[bankStart];
-	memory->currentBank = bank;
+	gb->memory.romBank = &gb->memory.rom[bankStart];
+	gb->memory.currentBank = bank;
+	if (gb->cpu->pc < GB_BASE_VRAM) {
+		gb->cpu->memory.setActiveRegion(gb->cpu, gb->cpu->pc);
+	}
+}
+
+void GBMBCSwitchBank0(struct GB* gb, int bank) {
+	size_t bankStart = bank * GB_SIZE_CART_BANK0 << gb->memory.mbcState.mbc1.multicartStride;
+	if (bankStart + GB_SIZE_CART_BANK0 > gb->memory.romSize) {
+		mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid ROM bank: %0X", bank);
+		bankStart &= (gb->memory.romSize - 1);
+	}
+	gb->memory.romBase = &gb->memory.rom[bankStart];
+	if (gb->cpu->pc < GB_SIZE_CART_BANK0) {
+		gb->cpu->memory.setActiveRegion(gb->cpu, gb->cpu->pc);
+	}
+}
+
+static bool _isMulticart(const uint8_t* mem) {
+	bool success = true;
+	struct VFile* vf;
+
+	vf = VFileFromConstMemory(&mem[GB_SIZE_CART_BANK0 * 0x10], 1024);
+	success = success && GBIsROM(vf);
+	vf->close(vf);
+
+	vf = VFileFromConstMemory(&mem[GB_SIZE_CART_BANK0 * 0x20], 1024);
+	success = success && GBIsROM(vf);
+	vf->close(vf);
+
+	return success;
 }
 
 void GBMBCSwitchSramBank(struct GB* gb, int bank) {
 	size_t bankStart = bank * GB_SIZE_EXTERNAL_RAM;
-	GBResizeSram(gb, (bank + 1) * GB_SIZE_EXTERNAL_RAM);
+	if (bankStart + GB_SIZE_EXTERNAL_RAM > gb->sramSize) {
+		mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid RAM bank: %0X", bank);
+		bankStart &= (gb->sramSize - 1);
+		bank = bankStart / GB_SIZE_EXTERNAL_RAM;
+	}
 	gb->memory.sramBank = &gb->memory.sram[bankStart];
 	gb->memory.sramCurrentBank = bank;
 }
@@ -62,6 +103,12 @@ void GBMBCInit(struct GB* gb) {
 		case 3:
 			gb->sramSize = 0x8000;
 			break;
+		case 4:
+			gb->sramSize = 0x20000;
+			break;
+		case 5:
+			gb->sramSize = 0x10000;
+			break;
 		}
 
 		if (gb->memory.mbcType == GB_MBC_AUTODETECT) {
@@ -75,6 +122,11 @@ void GBMBCInit(struct GB* gb) {
 			case 2:
 			case 3:
 				gb->memory.mbcType = GB_MBC1;
+				if (gb->memory.romSize >= GB_SIZE_CART_BANK0 * 0x31 && _isMulticart(gb->memory.rom)) {
+					gb->memory.mbcState.mbc1.multicartStride = 4;
+				} else {
+					gb->memory.mbcState.mbc1.multicartStride = 5;
+				}
 				break;
 			case 5:
 			case 6:
@@ -108,6 +160,12 @@ void GBMBCInit(struct GB* gb) {
 			case 0x22:
 				gb->memory.mbcType = GB_MBC7;
 				break;
+			case 0xFC:
+				gb->memory.mbcType = GB_POCKETCAM;
+				break;
+			case 0xFD:
+				gb->memory.mbcType = GB_HuC1;
+				break;
 			case 0xFE:
 				gb->memory.mbcType = GB_HuC3;
 				break;
@@ -116,54 +174,78 @@ void GBMBCInit(struct GB* gb) {
 	} else {
 		gb->memory.mbcType = GB_MBC_NONE;
 	}
+	gb->memory.mbcRead = NULL;
 	switch (gb->memory.mbcType) {
 	case GB_MBC_NONE:
-		gb->memory.mbc = _GBMBCNone;
+		gb->memory.mbcWrite = _GBMBCNone;
 		break;
 	case GB_MBC1:
-		gb->memory.mbc = _GBMBC1;
+		gb->memory.mbcWrite = _GBMBC1;
 		break;
 	case GB_MBC2:
-		gb->memory.mbc = _GBMBC2;
-		gb->sramSize = 0x200;
+		gb->memory.mbcWrite = _GBMBC2;
+		gb->memory.mbcRead = _GBMBC2Read;
+		gb->sramSize = 0x100;
 		break;
 	case GB_MBC3:
-		gb->memory.mbc = _GBMBC3;
+		gb->memory.mbcWrite = _GBMBC3;
 		break;
 	default:
 		mLOG(GB_MBC, WARN, "Unknown MBC type: %02X", cart->type);
 		// Fall through
 	case GB_MBC5:
-		gb->memory.mbc = _GBMBC5;
+		gb->memory.mbcWrite = _GBMBC5;
 		break;
 	case GB_MBC6:
 		mLOG(GB_MBC, WARN, "unimplemented MBC: MBC6");
-		gb->memory.mbc = _GBMBC6;
+		gb->memory.mbcWrite = _GBMBC6;
 		break;
 	case GB_MBC7:
-		gb->memory.mbc = _GBMBC7;
-		gb->sramSize = GB_SIZE_EXTERNAL_RAM;
+		gb->memory.mbcWrite = _GBMBC7;
+		gb->memory.mbcRead = _GBMBC7Read;
+		gb->sramSize = 0x100;
 		break;
 	case GB_MMM01:
 		mLOG(GB_MBC, WARN, "unimplemented MBC: MMM01");
-		gb->memory.mbc = _GBMBC1;
+		gb->memory.mbcWrite = _GBMBC1;
 		break;
 	case GB_HuC1:
 		mLOG(GB_MBC, WARN, "unimplemented MBC: HuC-1");
-		gb->memory.mbc = _GBMBC1;
+		gb->memory.mbcWrite = _GBMBC1;
 		break;
 	case GB_HuC3:
-		gb->memory.mbc = _GBHuC3;
+		gb->memory.mbcWrite = _GBHuC3;
 		break;
 	case GB_MBC3_RTC:
 		memset(gb->memory.rtcRegs, 0, sizeof(gb->memory.rtcRegs));
-		gb->memory.mbc = _GBMBC3;
+		gb->memory.mbcWrite = _GBMBC3;
 		break;
 	case GB_MBC5_RUMBLE:
-		gb->memory.mbc = _GBMBC5;
+		gb->memory.mbcWrite = _GBMBC5;
+		break;
+	case GB_POCKETCAM:
+		gb->memory.mbcWrite = _GBPocketCam;
+		gb->memory.mbcRead = _GBPocketCamRead;
 		break;
 	}
 
+	gb->memory.currentBank = 1;
+	gb->memory.sramCurrentBank = 0;
+	gb->memory.sramAccess = false;
+	gb->memory.rtcAccess = false;
+	gb->memory.activeRtcReg = 0;
+	gb->memory.rtcLatched = false;
+	gb->memory.rtcLastLatch = 0;
+	if (gb->memory.rtc) {
+		if (gb->memory.rtc->sample) {
+			gb->memory.rtc->sample(gb->memory.rtc);
+		}
+		gb->memory.rtcLastLatch = gb->memory.rtc->unixTime(gb->memory.rtc);
+	} else {
+		gb->memory.rtcLastLatch = time(0);
+	}
+	memset(&gb->memory.rtcRegs, 0, sizeof(gb->memory.rtcRegs));
+
 	GBResizeSram(gb, gb->sramSize);
 
 	if (gb->memory.mbcType == GB_MBC3_RTC) {
@@ -225,6 +307,7 @@ static void _latchRtc(struct mRTCSource* rtc, uint8_t* rtcRegs, time_t* rtcLastL
 void _GBMBC1(struct GB* gb, uint16_t address, uint8_t value) {
 	struct GBMemory* memory = &gb->memory;
 	int bank = value & 0x1F;
+	int stride = 1 << memory->mbcState.mbc1.multicartStride;
 	switch (address >> 13) {
 	case 0x0:
 		switch (value) {
@@ -245,21 +328,23 @@ void _GBMBC1(struct GB* gb, uint16_t address, uint8_t value) {
 		if (!bank) {
 			++bank;
 		}
-		GBMBCSwitchBank(memory, bank | (memory->currentBank & 0x60));
+		bank &= stride - 1;
+		GBMBCSwitchBank(gb, bank | (memory->currentBank & (3 * stride)));
 		break;
 	case 0x2:
 		bank &= 3;
-		if (!memory->mbcState.mbc1.mode) {
-			GBMBCSwitchBank(memory, (bank << 5) | (memory->currentBank & 0x1F));
-		} else {
+		if (memory->mbcState.mbc1.mode) {
+			GBMBCSwitchBank0(gb, bank);
 			GBMBCSwitchSramBank(gb, bank);
 		}
+		GBMBCSwitchBank(gb, (bank << memory->mbcState.mbc1.multicartStride) | (memory->currentBank & (stride - 1)));
 		break;
 	case 0x3:
 		memory->mbcState.mbc1.mode = value & 1;
 		if (memory->mbcState.mbc1.mode) {
-			GBMBCSwitchBank(memory, memory->currentBank & 0x1F);
+			GBMBCSwitchBank0(gb, memory->currentBank >> memory->mbcState.mbc1.multicartStride);
 		} else {
+			GBMBCSwitchBank0(gb, 0);
 			GBMBCSwitchSramBank(gb, 0);
 		}
 		break;
@@ -272,6 +357,7 @@ void _GBMBC1(struct GB* gb, uint16_t address, uint8_t value) {
 
 void _GBMBC2(struct GB* gb, uint16_t address, uint8_t value) {
 	struct GBMemory* memory = &gb->memory;
+	int shift = (address & 1) * 4;
 	int bank = value & 0xF;
 	switch (address >> 13) {
 	case 0x0:
@@ -281,7 +367,6 @@ void _GBMBC2(struct GB* gb, uint16_t address, uint8_t value) {
 			break;
 		case 0xA:
 			memory->sramAccess = true;
-			GBMBCSwitchSramBank(gb, memory->sramCurrentBank);
 			break;
 		default:
 			// TODO
@@ -293,13 +378,27 @@ void _GBMBC2(struct GB* gb, uint16_t address, uint8_t value) {
 		if (!bank) {
 			++bank;
 		}
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
+	case 0x5:
+		if (!memory->sramAccess) {
+			return;
+		}
+		address &= 0x1FF;
+		memory->sramBank[(address >> 1)] &= 0xF0 >> shift;
+		memory->sramBank[(address >> 1)] |= (value & 0xF) << shift;
 	default:
 		// TODO
 		mLOG(GB_MBC, STUB, "MBC2 unknown address: %04X:%02X", address, value);
 		break;
-	}}
+	}
+}
+
+static uint8_t _GBMBC2Read(struct GBMemory* memory, uint16_t address) {
+	address &= 0x1FF;
+	int shift = (address & 1) * 4;
+	return (memory->sramBank[(address >> 1)] >> shift) | 0xF0;
+}
 
 void _GBMBC3(struct GB* gb, uint16_t address, uint8_t value) {
 	struct GBMemory* memory = &gb->memory;
@@ -324,7 +423,7 @@ void _GBMBC3(struct GB* gb, uint16_t address, uint8_t value) {
 		if (!bank) {
 			++bank;
 		}
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x2:
 		if (value < 4) {
@@ -368,11 +467,11 @@ void _GBMBC5(struct GB* gb, uint16_t address, uint8_t value) {
 		break;
 	case 0x2:
 		bank = (memory->currentBank & 0x100) | value;
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x3:
 		bank = (memory->currentBank & 0xFF) | ((value & 1) << 8);
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x4:
 	case 0x5:
@@ -398,15 +497,27 @@ void _GBMBC6(struct GB* gb, uint16_t address, uint8_t value) {
 }
 
 void _GBMBC7(struct GB* gb, uint16_t address, uint8_t value) {
-	struct GBMemory* memory = &gb->memory;
 	int bank = value & 0x7F;
 	switch (address >> 13) {
+	case 0x0:
+		switch (value) {
+		default:
+		case 0:
+			gb->memory.mbcState.mbc7.access = 0;
+			break;
+		case 0xA:
+			gb->memory.mbcState.mbc7.access |= 1;
+			break;
+		}
+		break;
 	case 0x1:
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x2:
-		if (value < 0x10) {
-			GBMBCSwitchSramBank(gb, value);
+		if (value == 0x40) {
+			gb->memory.mbcState.mbc7.access |= 2;
+		} else {
+			gb->memory.mbcState.mbc7.access &= ~2;
 		}
 		break;
 	default:
@@ -416,19 +527,17 @@ void _GBMBC7(struct GB* gb, uint16_t address, uint8_t value) {
 	}
 }
 
-uint8_t GBMBC7Read(struct GBMemory* memory, uint16_t address) {
+uint8_t _GBMBC7Read(struct GBMemory* memory, uint16_t address) {
 	struct GBMBC7State* mbc7 = &memory->mbcState.mbc7;
+	if (mbc7->access != 3) {
+		return 0xFF;
+	}
 	switch (address & 0xF0) {
-	case 0x00:
-	case 0x10:
-	case 0x60:
-	case 0x70:
-		return 0;
 	case 0x20:
 		if (memory->rotation && memory->rotation->readTiltX) {
 			int32_t x = -memory->rotation->readTiltX(memory->rotation);
 			x >>= 21;
-			x += 2047;
+			x += 0x81D0;
 			return x;
 		}
 		return 0xFF;
@@ -436,7 +545,7 @@ uint8_t GBMBC7Read(struct GBMemory* memory, uint16_t address) {
 		if (memory->rotation && memory->rotation->readTiltX) {
 			int32_t x = -memory->rotation->readTiltX(memory->rotation);
 			x >>= 21;
-			x += 2047;
+			x += 0x81D0;
 			return x >> 8;
 		}
 		return 7;
@@ -444,7 +553,7 @@ uint8_t GBMBC7Read(struct GBMemory* memory, uint16_t address) {
 		if (memory->rotation && memory->rotation->readTiltY) {
 			int32_t y = -memory->rotation->readTiltY(memory->rotation);
 			y >>= 21;
-			y += 2047;
+			y += 0x81D0;
 			return y;
 		}
 		return 0xFF;
@@ -452,144 +561,142 @@ uint8_t GBMBC7Read(struct GBMemory* memory, uint16_t address) {
 		if (memory->rotation && memory->rotation->readTiltY) {
 			int32_t y = -memory->rotation->readTiltY(memory->rotation);
 			y >>= 21;
-			y += 2047;
+			y += 0x81D0;
 			return y >> 8;
 		}
 		return 7;
+	case 0x60:
+		return 0;
 	case 0x80:
-		return (mbc7->sr >> 16) & 1;
+		return mbc7->eeprom;
 	default:
 		return 0xFF;
 	}
 }
 
 void GBMBC7Write(struct GBMemory* memory, uint16_t address, uint8_t value) {
-	if ((address & 0xF0) != 0x80) {
+	struct GBMBC7State* mbc7 = &memory->mbcState.mbc7;
+	if (mbc7->access != 3) {
 		return;
 	}
-	struct GBMBC7State* mbc7 = &memory->mbcState.mbc7;
-	GBMBC7Field old = memory->mbcState.mbc7.field;
-	mbc7->field = GBMBC7FieldClearIO(value);
-	if (!GBMBC7FieldIsCS(old) && GBMBC7FieldIsCS(value)) {
-		if (mbc7->state == GBMBC7_STATE_WRITE) {
-			if (mbc7->writable) {
-				memory->sramBank[mbc7->address * 2] = mbc7->sr >> 8;
-				memory->sramBank[mbc7->address * 2 + 1] = mbc7->sr;
-			}
-			mbc7->sr = 0x1FFFF;
-			mbc7->state = GBMBC7_STATE_NULL;
-		} else {
-			mbc7->state = GBMBC7_STATE_IDLE;
+	switch (address & 0xF0) {
+	case 0x00:
+		mbc7->latch = (value & 0x55) == 0x55;
+		return;
+	case 0x10:
+		mbc7->latch |= (value & 0xAA);
+		if (mbc7->latch == 0xAB && memory->rotation && memory->rotation->sample) {
+			memory->rotation->sample(memory->rotation);
 		}
+		mbc7->latch = 0;
+		return;
+	default:
+		mLOG(GB_MBC, STUB, "MBC7 unknown register: %04X:%02X", address, value);
+		return;
+	case 0x80:
+		break;
 	}
-	if (!GBMBC7FieldIsSK(old) && GBMBC7FieldIsSK(value)) {
-		if (mbc7->state > GBMBC7_STATE_IDLE && mbc7->state != GBMBC7_STATE_READ) {
+	GBMBC7Field old = memory->mbcState.mbc7.eeprom;
+	value = GBMBC7FieldFillDO(value); // Hi-Z
+	if (!GBMBC7FieldIsCS(old) && GBMBC7FieldIsCS(value)) {
+		mbc7->state = GBMBC7_STATE_IDLE;
+	}
+	if (!GBMBC7FieldIsCLK(old) && GBMBC7FieldIsCLK(value)) {
+		if (mbc7->state == GBMBC7_STATE_READ_COMMAND || mbc7->state == GBMBC7_STATE_EEPROM_WRITE || mbc7->state == GBMBC7_STATE_EEPROM_WRAL) {
 			mbc7->sr <<= 1;
-			mbc7->sr |= GBMBC7FieldGetIO(value);
+			mbc7->sr |= GBMBC7FieldGetDI(value);
 			++mbc7->srBits;
 		}
 		switch (mbc7->state) {
 		case GBMBC7_STATE_IDLE:
-			if (GBMBC7FieldIsIO(value)) {
+			if (GBMBC7FieldIsDI(value)) {
 				mbc7->state = GBMBC7_STATE_READ_COMMAND;
 				mbc7->srBits = 0;
 				mbc7->sr = 0;
 			}
 			break;
 		case GBMBC7_STATE_READ_COMMAND:
-			if (mbc7->srBits == 2) {
-				mbc7->state = GBMBC7_STATE_READ_ADDRESS;
-				mbc7->srBits = 0;
-				mbc7->command = mbc7->sr;
-			}
-			break;
-		case GBMBC7_STATE_READ_ADDRESS:
-			if (mbc7->srBits == 8) {
-				mbc7->state = GBMBC7_STATE_COMMAND_0 + mbc7->command;
-				mbc7->srBits = 0;
-				mbc7->address = mbc7->sr;
-				if (mbc7->state == GBMBC7_STATE_COMMAND_0) {
-					switch (mbc7->address >> 6) {
-					case 0:
-						mbc7->writable = false;
-						mbc7->state = GBMBC7_STATE_NULL;
-						break;
-					case 3:
-						mbc7->writable = true;
-						mbc7->state = GBMBC7_STATE_NULL;
-						break;
-					}
+			if (mbc7->srBits == 10) {
+				mbc7->state = 0x10 | (mbc7->sr >> 6);
+				if (mbc7->state & 0xC) {
+					mbc7->state &= ~0x3;
 				}
-			}
-			break;
-		case GBMBC7_STATE_COMMAND_0:
-			if (mbc7->srBits == 16) {
-				switch (mbc7->address >> 6) {
-				case 0:
-					mbc7->writable = false;
-					mbc7->state = GBMBC7_STATE_NULL;
-					break;
-				case 1:
-					mbc7->state = GBMBC7_STATE_WRITE;
-					if (mbc7->writable) {
-						int i;
-						for (i = 0; i < 256; ++i) {
-							memory->sramBank[i * 2] = mbc7->sr >> 8;
-							memory->sramBank[i * 2 + 1] = mbc7->sr;
-						}
-					}
-					break;
-				case 2:
-					mbc7->state = GBMBC7_STATE_WRITE;
-					if (mbc7->writable) {
-						int i;
-						for (i = 0; i < 256; ++i) {
-							memory->sramBank[i * 2] = 0xFF;
-							memory->sramBank[i * 2 + 1] = 0xFF;
-						}
-					}
-					break;
-				case 3:
-					mbc7->writable = true;
-					mbc7->state = GBMBC7_STATE_NULL;
-					break;
-				}
-			}
-			break;
-		case GBMBC7_STATE_COMMAND_SR_WRITE:
-			if (mbc7->srBits == 16) {
 				mbc7->srBits = 0;
-				mbc7->state = GBMBC7_STATE_WRITE;
+				mbc7->address = mbc7->sr & 0x7F;
 			}
 			break;
-		case GBMBC7_STATE_COMMAND_SR_READ:
-			if (mbc7->srBits == 1) {
-				mbc7->sr = memory->sramBank[mbc7->address * 2] << 8;
-				mbc7->sr |= memory->sramBank[mbc7->address * 2 + 1];
-				mbc7->srBits = 0;
-				mbc7->state = GBMBC7_STATE_READ;
-			}
-			break;
-		case GBMBC7_STATE_COMMAND_SR_FILL:
-			if (mbc7->srBits == 16) {
-				mbc7->sr = 0xFFFF;
-				mbc7->srBits = 0;
-				mbc7->state = GBMBC7_STATE_WRITE;
+		case GBMBC7_STATE_DO:
+			value = GBMBC7FieldSetDO(value, mbc7->sr >> 15);
+			mbc7->sr <<= 1;
+			--mbc7->srBits;
+			if (!mbc7->srBits) {
+				mbc7->state = GBMBC7_STATE_IDLE;
 			}
 			break;
 		default:
 			break;
 		}
-	} else if (GBMBC7FieldIsSK(old) && !GBMBC7FieldIsSK(value)) {
-		if (mbc7->state == GBMBC7_STATE_READ) {
-			mbc7->sr <<= 1;
-			++mbc7->srBits;
+		switch (mbc7->state) {
+		case GBMBC7_STATE_EEPROM_EWEN:
+			mbc7->writable = true;
+			mbc7->state = GBMBC7_STATE_IDLE;
+			break;
+		case GBMBC7_STATE_EEPROM_EWDS:
+			mbc7->writable = false;
+			mbc7->state = GBMBC7_STATE_IDLE;
+			break;
+		case GBMBC7_STATE_EEPROM_WRITE:
 			if (mbc7->srBits == 16) {
-				mbc7->srBits = 0;
-				mbc7->state = GBMBC7_STATE_NULL;
+				if (mbc7->writable) {
+					memory->sram[mbc7->address * 2] = mbc7->sr >> 8;
+					memory->sram[mbc7->address * 2 + 1] = mbc7->sr;
+				}
+				mbc7->state = GBMBC7_STATE_IDLE;
 			}
+			break;
+		case GBMBC7_STATE_EEPROM_ERASE:
+			if (mbc7->writable) {
+				memory->sram[mbc7->address * 2] = 0xFF;
+				memory->sram[mbc7->address * 2 + 1] = 0xFF;
+			}
+			mbc7->state = GBMBC7_STATE_IDLE;
+			break;
+		case GBMBC7_STATE_EEPROM_READ:
+			mbc7->srBits = 16;
+			mbc7->sr = memory->sram[mbc7->address * 2] << 8;
+			mbc7->sr |= memory->sram[mbc7->address * 2 + 1];
+			mbc7->state = GBMBC7_STATE_DO;
+			value = GBMBC7FieldClearDO(value);
+			break;
+		case GBMBC7_STATE_EEPROM_WRAL:
+			if (mbc7->srBits == 16) {
+				if (mbc7->writable) {
+					int i;
+					for (i = 0; i < 128; ++i) {
+						memory->sram[i * 2] = mbc7->sr >> 8;
+						memory->sram[i * 2 + 1] = mbc7->sr;
+					}
+				}
+				mbc7->state = GBMBC7_STATE_IDLE;
+			}
+			break;
+		case GBMBC7_STATE_EEPROM_ERAL:
+			if (mbc7->writable) {
+				int i;
+				for (i = 0; i < 128; ++i) {
+					memory->sram[i * 2] = 0xFF;
+					memory->sram[i * 2 + 1] = 0xFF;
+				}
+			}
+			mbc7->state = GBMBC7_STATE_IDLE;
+			break;
+		default:
+			break;
 		}
+	} else if (GBMBC7FieldIsCS(value) && GBMBC7FieldIsCLK(old) && !GBMBC7FieldIsCLK(value)) {
+		value = GBMBC7FieldSetDO(value, GBMBC7FieldGetDO(old));
 	}
+	mbc7->eeprom = value;
 }
 
 void _GBHuC3(struct GB* gb, uint16_t address, uint8_t value) {
@@ -612,7 +719,7 @@ void _GBHuC3(struct GB* gb, uint16_t address, uint8_t value) {
 		}
 		break;
 	case 0x1:
-		GBMBCSwitchBank(memory, bank);
+		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x2:
 		GBMBCSwitchSramBank(gb, bank);
@@ -624,23 +731,59 @@ void _GBHuC3(struct GB* gb, uint16_t address, uint8_t value) {
 	}
 }
 
+void _GBPocketCam(struct GB* gb, uint16_t address, uint8_t value) {
+	struct GBMemory* memory = &gb->memory;
+	int bank = value & 0x3F;
+	switch (address >> 13) {
+	case 0x0:
+		switch (value) {
+		case 0:
+			memory->sramAccess = false;
+			break;
+		case 0xA:
+			memory->sramAccess = true;
+			GBMBCSwitchSramBank(gb, memory->sramCurrentBank);
+			break;
+		default:
+			// TODO
+			mLOG(GB_MBC, STUB, "Pocket Cam unknown value %02X", value);
+			break;
+		}
+		break;
+	case 0x1:
+		GBMBCSwitchBank(gb, bank);
+		break;
+	case 0x2:
+		if (value < 0x10) {
+			GBMBCSwitchSramBank(gb, value);
+			memory->mbcState.pocketCam.registersActive = false;
+		} else {
+			memory->mbcState.pocketCam.registersActive = true;
+		}
+		break;
+	default:
+		mLOG(GB_MBC, STUB, "Pocket Cam unknown address: %04X:%02X", address, value);
+		break;
+	}
+}
+
+uint8_t _GBPocketCamRead(struct GBMemory* memory, uint16_t address) {
+	if (memory->mbcState.pocketCam.registersActive) {
+		return 0;
+	}
+	return memory->sramBank[address & (GB_SIZE_EXTERNAL_RAM - 1)];
+}
+
 void GBMBCRTCRead(struct GB* gb) {
 	struct GBMBCRTCSaveBuffer rtcBuffer;
 	struct VFile* vf = gb->sramVf;
 	if (!vf) {
 		return;
 	}
-	ssize_t end = vf->seek(vf, -sizeof(rtcBuffer), SEEK_END);
-	switch (end & 0x1FFF) {
-	case 0:
-		break;
-	case 0x1FFC:
-		vf->seek(vf, -sizeof(rtcBuffer) - 4, SEEK_END);
-		break;
-	default:
+	vf->seek(vf, gb->sramSize, SEEK_SET);
+	if (vf->read(vf, &rtcBuffer, sizeof(rtcBuffer)) < (ssize_t) sizeof(rtcBuffer) - 4) {
 		return;
 	}
-	vf->read(vf, &rtcBuffer, sizeof(rtcBuffer));
 
 	LOAD_32LE(gb->memory.rtcRegs[0], 0, &rtcBuffer.latchedSec);
 	LOAD_32LE(gb->memory.rtcRegs[1], 0, &rtcBuffer.latchedMin);
@@ -672,9 +815,9 @@ void GBMBCRTCWrite(struct GB* gb) {
 	STORE_32LE(gb->memory.rtcRegs[2], 0, &rtcBuffer.latchedHour);
 	STORE_32LE(gb->memory.rtcRegs[3], 0, &rtcBuffer.latchedDays);
 	STORE_32LE(gb->memory.rtcRegs[4], 0, &rtcBuffer.latchedDaysHi);
-	STORE_64LE(rtcLastLatch, 0, &rtcBuffer.unixTime);
+	STORE_64LE(gb->memory.rtcLastLatch, 0, &rtcBuffer.unixTime);
 
-	if (vf->size(vf) == gb->sramSize) {
+	if ((size_t) vf->size(vf) < gb->sramSize + sizeof(rtcBuffer)) {
 		// Writing past the end of the file can invalidate the file mapping
 		vf->unmap(vf, gb->memory.sram, gb->sramSize);
 		gb->memory.sram = NULL;
