@@ -3,11 +3,11 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "gameshark.h"
+#include <mgba/internal/gba/cheats.h>
 
 #include "gba/cheats/parv3.h"
-#include "gba/gba.h"
-#include "util/string.h"
+#include <mgba/internal/gba/gba.h>
+#include <mgba-util/string.h>
 
 const uint32_t GBACheatGameSharkSeeds[4] = { 0x09F4FBBD, 0x9681884A, 0x352027E9, 0xF3DEE5A7 };
 
@@ -90,6 +90,7 @@ void GBACheatSetGameSharkVersion(struct GBACheatSet* cheats, int version) {
 bool GBACheatAddGameSharkRaw(struct GBACheatSet* cheats, uint32_t op1, uint32_t op2) {
 	enum GBAGameSharkType type = op1 >> 28;
 	struct mCheat* cheat = 0;
+	int romPatch = 0;
 
 	if (cheats->incompleteCheat != COMPLETE) {
 		struct mCheat* incompleteCheat = mCheatListGetPointer(&cheats->d.list, cheats->incompleteCheat);
@@ -145,10 +146,16 @@ bool GBACheatAddGameSharkRaw(struct GBACheatSet* cheats, uint32_t op1, uint32_t
 		cheats->incompleteCheat = mCheatListIndex(&cheats->d.list, cheat);
 		break;
 	case GSA_PATCH:
-		cheats->romPatches[0].address = BASE_CART0 | ((op1 & 0xFFFFFF) << 1);
-		cheats->romPatches[0].newValue = op2;
-		cheats->romPatches[0].applied = false;
-		cheats->romPatches[0].exists = true;
+		while (cheats->romPatches[romPatch].exists) {
+			++romPatch;
+			if (romPatch >= MAX_ROM_PATCHES) {
+				break;
+			}
+		}
+		cheats->romPatches[romPatch].address = BASE_CART0 | ((op1 & 0xFFFFFF) << 1);
+		cheats->romPatches[romPatch].newValue = op2;
+		cheats->romPatches[romPatch].applied = false;
+		cheats->romPatches[romPatch].exists = true;
 		return true;
 	case GSA_BUTTON:
 		// TODO: Implement button
@@ -171,6 +178,7 @@ bool GBACheatAddGameSharkRaw(struct GBACheatSet* cheats, uint32_t op1, uint32_t
 		cheat->address = op2 & 0x0FFFFFFF;
 		cheat->operand = op1 & 0xFFFF;
 		cheat->repeat = (op1 >> 16) & 0xFF;
+		cheat->negativeRepeat = 0;
 		return true;
 	case GSA_HOOK:
 		if (cheats->hook) {
@@ -187,6 +195,7 @@ bool GBACheatAddGameSharkRaw(struct GBACheatSet* cheats, uint32_t op1, uint32_t
 	}
 	cheat->operand = op2;
 	cheat->repeat = 1;
+	cheat->negativeRepeat = 0;
 	return true;
 }
 
@@ -203,8 +212,9 @@ bool GBACheatAddGameShark(struct GBACheatSet* set, uint32_t op1, uint32_t op2) {
 		GBACheatSetGameSharkVersion(set, 1);
 	// Fall through
 	case 1:
-	case 2:
 		GBACheatDecryptGameShark(&o1, &o2, set->gsaSeeds);
+	// Fall through
+	case 2:
 		return GBACheatAddGameSharkRaw(set, o1, o2);
 	}
 	return false;
@@ -226,3 +236,63 @@ bool GBACheatAddGameSharkLine(struct GBACheatSet* cheats, const char* line) {
 	}
 	return GBACheatAddGameShark(cheats, op1, op2);
 }
+
+int GBACheatGameSharkProbability(uint32_t op1, uint32_t op2) {
+	int probability = 0;
+	if (op2 == 0x001DC0DE) {
+		return 0x100;
+	}
+	uint32_t address = op1 & 0x0FFFFFFF;
+	switch (op1 >> 28) {
+	case GSA_ASSIGN_1:
+		probability += 0x20;
+		if (op2 & 0xFFFFFF00) {
+			probability -= 0x10;
+		}
+		probability += GBACheatAddressIsReal(address);
+		break;
+	case GSA_ASSIGN_2:
+		probability += 0x20;
+		if (op2 & 0xFFFF0000) {
+			probability -= 0x10;
+		}
+		probability += GBACheatAddressIsReal(address);
+		break;
+	case GSA_ASSIGN_4:
+		probability += 0x20;
+		probability += GBACheatAddressIsReal(address);
+		break;
+	case GSA_PATCH:
+		probability += 0x20;
+		if (op2 & 0xCFFF0000) {
+			probability -= 0x10;
+		}
+		break;
+	case GSA_BUTTON:
+		probability += 0x10;
+		break;
+	case GSA_IF_EQ:
+		probability += 0x20;
+		if (op2 & 0xFFFF0000) {
+			probability -= 0x10;
+		}
+		probability += GBACheatAddressIsReal(address);
+		break;
+	case GSA_IF_EQ_RANGE:
+		probability += 0x20;
+		probability += GBACheatAddressIsReal(op2);
+		if (op1 & 0x0F000000) {
+			probability -= 0x10;
+		}
+		break;
+	case GSA_HOOK:
+		probability += 0x20;
+		if (op2 & 0xFFFF0000) {
+			probability -= 0x10;
+		}
+		break;
+	default:
+		probability -= 0x40;
+	}
+	return probability;
+}
