@@ -3,14 +3,14 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "software-private.h"
+#include "gba/renderers/software-private.h"
 
-#include "gba/gba.h"
-#include "gba/io.h"
-#include "gba/renderers/tile-cache.h"
+#include <mgba/core/tile-cache.h>
+#include <mgba/internal/arm/macros.h>
+#include <mgba/internal/gba/io.h>
 
-#include "util/arm-algo.h"
-#include "util/memory.h"
+#include <mgba-util/arm-algo.h>
+#include <mgba-util/memory.h>
 
 static void GBAVideoSoftwareRendererInit(struct GBAVideoRenderer* renderer);
 static void GBAVideoSoftwareRendererDeinit(struct GBAVideoRenderer* renderer);
@@ -97,6 +97,7 @@ static void GBAVideoSoftwareRendererReset(struct GBAVideoRenderer* renderer) {
 		LOAD_16(entry, i, softwareRenderer->d.palette);
 		GBAVideoSoftwareRendererWritePalette(renderer, i, entry);
 	}
+	softwareRenderer->blendDirty = false;
 	_updatePalettes(softwareRenderer);
 
 	softwareRenderer->blda = 0;
@@ -146,6 +147,7 @@ static uint16_t GBAVideoSoftwareRendererWriteVideoRegister(struct GBAVideoRender
 	struct GBAVideoSoftwareRenderer* softwareRenderer = (struct GBAVideoSoftwareRenderer*) renderer;
 	switch (address) {
 	case REG_DISPCNT:
+		value &= 0xFFF7;
 		softwareRenderer->dispcnt = value;
 		GBAVideoSoftwareRendererUpdateDISPCNT(softwareRenderer);
 		break;
@@ -261,11 +263,14 @@ static uint16_t GBAVideoSoftwareRendererWriteVideoRegister(struct GBAVideoRender
 		value &= 0x1F1F;
 		break;
 	case REG_BLDY:
-		softwareRenderer->bldy = value & 0x1F;
-		if (softwareRenderer->bldy > 0x10) {
-			softwareRenderer->bldy = 0x10;
+		value &= 0x1F;
+		if (value > 0x10) {
+			value = 0x10;
+		}
+		if (softwareRenderer->bldy != value) {
+			softwareRenderer->bldy = value;
+			softwareRenderer->blendDirty = true;
 		}
-		_updatePalettes(softwareRenderer);
 		break;
 	case REG_WIN0H:
 		softwareRenderer->winN[0].h.end = value;
@@ -343,7 +348,7 @@ static uint16_t GBAVideoSoftwareRendererWriteVideoRegister(struct GBAVideoRender
 
 static void GBAVideoSoftwareRendererWriteVRAM(struct GBAVideoRenderer* renderer, uint32_t address) {
 	if (renderer->cache) {
-		GBAVideoTileCacheWriteVRAM(renderer->cache, address);
+		mTileCacheWriteVRAM(renderer->cache, address);
 	}
 }
 
@@ -369,6 +374,7 @@ static void GBAVideoSoftwareRendererWritePalette(struct GBAVideoRenderer* render
 	color |= (value << 3) & 0xF8;
 	color |= (value << 6) & 0xF800;
 	color |= (value << 9) & 0xF80000;
+	color |= (color >> 5) & 0x070707;
 #endif
 	softwareRenderer->normalPalette[address >> 1] = color;
 	if (softwareRenderer->blendEffect == BLEND_BRIGHTEN) {
@@ -377,7 +383,7 @@ static void GBAVideoSoftwareRendererWritePalette(struct GBAVideoRenderer* render
 		softwareRenderer->variantPalette[address >> 1] = _darken(color, softwareRenderer->bldy);
 	}
 	if (renderer->cache) {
-		GBAVideoTileCacheWritePalette(renderer->cache, address);
+		mTileCacheWritePalette(renderer->cache, address);
 	}
 }
 
@@ -457,7 +463,6 @@ static void _breakWindowInner(struct GBAVideoSoftwareRenderer* softwareRenderer,
 static void _cleanOAM(struct GBAVideoSoftwareRenderer* renderer) {
 	int i;
 	int oamMax = 0;
-	int objwinMax = 0;
 	for (i = 0; i < 128; ++i) {
 		struct GBAObj obj;
 		LOAD_16(obj.a, 0, &renderer->d.oam->obj[i].a);
@@ -469,22 +474,14 @@ static void _cleanOAM(struct GBAVideoSoftwareRenderer* renderer) {
 				height <<= GBAObjAttributesAGetDoubleSize(obj.a);
 			}
 			if (GBAObjAttributesAGetY(obj.a) < VIDEO_VERTICAL_PIXELS || GBAObjAttributesAGetY(obj.a) + height >= VIDEO_VERTICAL_TOTAL_PIXELS) {
-				if (GBAObjAttributesAGetMode(obj.a) == OBJ_MODE_OBJWIN) {
-					renderer->objwinSprites[objwinMax].y = GBAObjAttributesAGetY(obj.a);
-					renderer->objwinSprites[objwinMax].endY = GBAObjAttributesAGetY(obj.a) + height;
-					renderer->objwinSprites[objwinMax].obj = obj;
-					++objwinMax;
-				} else {
-					renderer->sprites[oamMax].y = GBAObjAttributesAGetY(obj.a);
-					renderer->sprites[oamMax].endY = GBAObjAttributesAGetY(obj.a) + height;
-					renderer->sprites[oamMax].obj = obj;
-					++oamMax;
-				}
+				renderer->sprites[oamMax].y = GBAObjAttributesAGetY(obj.a);
+				renderer->sprites[oamMax].endY = GBAObjAttributesAGetY(obj.a) + height;
+				renderer->sprites[oamMax].obj = obj;
+				++oamMax;
 			}
 		}
 	}
 	renderer->oamMax = oamMax;
-	renderer->objwinMax = objwinMax;
 	renderer->oamDirty = 0;
 }
 
@@ -523,6 +520,10 @@ static void GBAVideoSoftwareRendererDrawScanline(struct GBAVideoRenderer* render
 	}
 
 	GBAVideoSoftwareRendererUpdateDISPCNT(softwareRenderer);
+	if (softwareRenderer->blendDirty) {
+		_updatePalettes(softwareRenderer);
+		softwareRenderer->blendDirty = false;
+	}
 
 	int w;
 	x = 0;
@@ -568,6 +569,14 @@ static void GBAVideoSoftwareRendererDrawScanline(struct GBAVideoRenderer* render
 	}
 	if (softwareRenderer->target1Obj && (softwareRenderer->blendEffect == BLEND_DARKEN || softwareRenderer->blendEffect == BLEND_BRIGHTEN)) {
 		x = 0;
+		uint32_t mask = FLAG_REBLEND | FLAG_TARGET_1 | FLAG_IS_BACKGROUND;
+		uint32_t match = FLAG_REBLEND;
+		if (GBARegisterDISPCNTIsObjwinEnable(softwareRenderer->dispcnt)) {
+			mask |= FLAG_OBJWIN;
+			if (GBAWindowControlIsBlendEnable(softwareRenderer->objwin.packed)) {
+				match |= FLAG_OBJWIN;
+			}
+		}
 		for (w = 0; w < softwareRenderer->nWindows; ++w) {
 			if (!GBAWindowControlIsBlendEnable(softwareRenderer->windows[w].control.packed)) {
 				continue;
@@ -576,14 +585,14 @@ static void GBAVideoSoftwareRendererDrawScanline(struct GBAVideoRenderer* render
 			if (softwareRenderer->blendEffect == BLEND_DARKEN) {
 				for (; x < end; ++x) {
 					uint32_t color = softwareRenderer->row[x];
-					if ((color & 0xFF000000) == FLAG_REBLEND) {
+					if ((color & mask) == match) {
 						softwareRenderer->row[x] = _darken(color, softwareRenderer->bldy);
 					}
 				}
 			} else if (softwareRenderer->blendEffect == BLEND_BRIGHTEN) {
 				for (; x < end; ++x) {
 					uint32_t color = softwareRenderer->row[x];
-					if ((color & 0xFF000000) == FLAG_REBLEND) {
+					if ((color & mask) == match) {
 						softwareRenderer->row[x] = _brighten(color, softwareRenderer->bldy);
 					}
 				}
@@ -710,7 +719,7 @@ static void GBAVideoSoftwareRendererWriteBLDCNT(struct GBAVideoSoftwareRenderer*
 	renderer->target2Bd = GBARegisterBLDCNTGetTarget2Bd(value);
 
 	if (oldEffect != renderer->blendEffect) {
-		_updatePalettes(renderer);
+		renderer->blendDirty = true;
 	}
 }
 
@@ -740,20 +749,7 @@ static void _drawScanline(struct GBAVideoSoftwareRenderer* renderer, int y) {
 			}
 			int i;
 			int drawn;
-			for (i = 0; i < renderer->objwinMax; ++i) {
-				int localY = y;
-				if (renderer->spriteCyclesRemaining <= 0) {
-					break;
-				}
-				struct GBAVideoSoftwareSprite* sprite = &renderer->objwinSprites[i];
-				if (GBAObjAttributesAIsMosaic(sprite->obj.a)) {
-					localY = mosaicY;
-				}
-				if ((localY < sprite->y && (sprite->endY - 256 < 0 || localY >= sprite->endY - 256)) || localY >= sprite->endY) {
-					continue;
-				}
-				GBAVideoSoftwareRendererPreprocessSprite(renderer, &sprite->obj, localY);
-			}
+
 			for (i = 0; i < renderer->oamMax; ++i) {
 				int localY = y;
 				if (renderer->spriteCyclesRemaining <= 0) {
@@ -820,10 +816,12 @@ static void _drawScanline(struct GBAVideoSoftwareRenderer* renderer, int y) {
 			}
 		}
 	}
-	renderer->bg[2].sx += renderer->bg[2].dmx;
-	renderer->bg[2].sy += renderer->bg[2].dmy;
-	renderer->bg[3].sx += renderer->bg[3].dmx;
-	renderer->bg[3].sy += renderer->bg[3].dmy;
+	if (GBARegisterDISPCNTGetMode(renderer->dispcnt) != 0) {
+		renderer->bg[2].sx += renderer->bg[2].dmx;
+		renderer->bg[2].sy += renderer->bg[2].dmy;
+		renderer->bg[3].sx += renderer->bg[3].dmx;
+		renderer->bg[3].sy += renderer->bg[3].dmy;
+	}
 }
 
 static void _updatePalettes(struct GBAVideoSoftwareRenderer* renderer) {
