@@ -3,13 +3,14 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "savedata.h"
+#include <mgba/internal/gba/savedata.h>
 
-#include "gba/gba.h"
-#include "gba/serialize.h"
+#include <mgba/internal/arm/macros.h>
+#include <mgba/internal/gba/gba.h>
+#include <mgba/internal/gba/serialize.h>
 
-#include "util/memory.h"
-#include "util/vfs.h"
+#include <mgba-util/memory.h>
+#include <mgba-util/vfs.h>
 
 #include <errno.h>
 #include <fcntl.h>
@@ -20,17 +21,24 @@
 // Other games vary from very little, with a fairly solid 20500 cycle count. (Observed on a SST (D4BF) chip).
 // An average estimation is as follows.
 #define FLASH_ERASE_CYCLES 30000
-#define FLASH_PROGRAM_CYCLES 18000
+#define FLASH_PROGRAM_CYCLES 650
 // This needs real testing, and is only an estimation currently
-#define EEPROM_SETTLE_CYCLES 1450
+#define EEPROM_SETTLE_CYCLES 115000
 #define CLEANUP_THRESHOLD 15
 
-mLOG_DEFINE_CATEGORY(GBA_SAVE, "GBA Savedata");
+mLOG_DEFINE_CATEGORY(GBA_SAVE, "GBA Savedata", "gba.savedata");
 
 static void _flashSwitchBank(struct GBASavedata* savedata, int bank);
 static void _flashErase(struct GBASavedata* savedata);
 static void _flashEraseSector(struct GBASavedata* savedata, uint16_t sectorStart);
 
+static void _ashesToAshes(struct mTiming* timing, void* user, uint32_t cyclesLate) {
+	UNUSED(timing);
+	UNUSED(user);
+	UNUSED(cyclesLate);
+	// Funk to funky
+}
+
 void GBASavedataInit(struct GBASavedata* savedata, struct VFile* vf) {
 	savedata->type = SAVEDATA_AUTODETECT;
 	savedata->data = 0;
@@ -41,6 +49,10 @@ void GBASavedataInit(struct GBASavedata* savedata, struct VFile* vf) {
 	savedata->mapMode = MAP_WRITE;
 	savedata->dirty = 0;
 	savedata->dirtAge = 0;
+	savedata->dust.name = "GBA Savedata Settling";
+	savedata->dust.priority = 0x70;
+	savedata->dust.context = savedata;
+	savedata->dust.callback = _ashesToAshes;
 }
 
 void GBASavedataDeinit(struct GBASavedata* savedata) {
@@ -135,7 +147,7 @@ size_t GBASavedataSize(struct GBASavedata* savedata) {
 	case SAVEDATA_FLASH1M:
 		return SIZE_CART_FLASH1M;
 	case SAVEDATA_EEPROM:
-		return SIZE_CART_EEPROM;
+		return (savedata->vf && savedata->vf->size(savedata->vf) == SIZE_CART_EEPROM512) ? SIZE_CART_EEPROM512 : SIZE_CART_EEPROM;
 	case SAVEDATA_FORCE_NONE:
 		return 0;
 	case SAVEDATA_AUTODETECT:
@@ -148,39 +160,35 @@ size_t GBASavedataSize(struct GBASavedata* savedata) {
 }
 
 bool GBASavedataLoad(struct GBASavedata* savedata, struct VFile* in) {
-	if (savedata->vf) {
+	if (savedata->data) {
+		if (!in && savedata->type != SAVEDATA_FORCE_NONE) {
+			return false;
+		}
+		ssize_t size = GBASavedataSize(savedata);
+		in->seek(in, 0, SEEK_SET);
+		return in->read(in, savedata->data, size) == size;
+	} else if (savedata->vf) {
 		off_t read = 0;
 		uint8_t buffer[2048];
-		memset(buffer, 0xFF, sizeof(buffer));
-		savedata->vf->seek(savedata->vf, 0, SEEK_SET);
-		while (savedata->vf->seek(savedata->vf, 0, SEEK_CUR) < savedata->vf->size(savedata->vf)) {
-			savedata->vf->write(savedata->vf, buffer, sizeof(buffer));
-		}
 		savedata->vf->seek(savedata->vf, 0, SEEK_SET);
 		if (in) {
+			in->seek(in, 0, SEEK_SET);
 			do {
 				read = in->read(in, buffer, sizeof(buffer));
 				read = savedata->vf->write(savedata->vf, buffer, read);
 			} while (read == sizeof(buffer));
 		}
+		memset(buffer, 0xFF, sizeof(buffer));
+		ssize_t fsize = savedata->vf->size(savedata->vf);
+		ssize_t pos = savedata->vf->seek(savedata->vf, 0, SEEK_CUR);
+		while (fsize - pos >= (ssize_t) sizeof(buffer)) {
+			savedata->vf->write(savedata->vf, buffer, sizeof(buffer));
+			pos = savedata->vf->seek(savedata->vf, 0, SEEK_CUR);
+		}
+		if (fsize - pos > 0) {
+			savedata->vf->write(savedata->vf, buffer, fsize - pos);
+		}
 		return read >= 0;
-	} else if (savedata->data) {
-		if (!in && savedata->type != SAVEDATA_FORCE_NONE) {
-			return false;
-		}
-		switch (savedata->type) {
-		case SAVEDATA_SRAM:
-			return in->read(in, savedata->data, SIZE_CART_SRAM) == SIZE_CART_SRAM;
-		case SAVEDATA_FLASH512:
-			return in->read(in, savedata->data, SIZE_CART_FLASH512) == SIZE_CART_FLASH512;
-		case SAVEDATA_FLASH1M:
-			return in->read(in, savedata->data, SIZE_CART_FLASH1M) == SIZE_CART_FLASH1M;
-		case SAVEDATA_EEPROM:
-			return in->read(in, savedata->data, SIZE_CART_EEPROM) == SIZE_CART_EEPROM;
-		case SAVEDATA_AUTODETECT:
-		case SAVEDATA_FORCE_NONE:
-			return true;
-		}
 	}
 	return true;
 }
@@ -236,7 +244,6 @@ void GBASavedataInitFlash(struct GBASavedata* savedata, bool realisticTiming) {
 	}
 
 	savedata->currentBank = savedata->data;
-	savedata->dust = 0;
 	savedata->realisticTiming = realisticTiming;
 	if (end < SIZE_CART_FLASH512) {
 		memset(&savedata->data[end], 0xFF, flashSize - end);
@@ -250,21 +257,23 @@ void GBASavedataInitEEPROM(struct GBASavedata* savedata, bool realisticTiming) {
 		mLOG(GBA_SAVE, WARN, "Can't re-initialize savedata");
 		return;
 	}
+	int32_t eepromSize = SIZE_CART_EEPROM512;
 	off_t end;
 	if (!savedata->vf) {
 		end = 0;
 		savedata->data = anonymousMemoryMap(SIZE_CART_EEPROM);
 	} else {
 		end = savedata->vf->size(savedata->vf);
-		if (end < SIZE_CART_EEPROM) {
-			savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM);
+		if (end < SIZE_CART_EEPROM512) {
+			savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM512);
+		} else if (end > SIZE_CART_EEPROM512) {
+			eepromSize = SIZE_CART_EEPROM;
 		}
-		savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_EEPROM, savedata->mapMode);
+		savedata->data = savedata->vf->map(savedata->vf, eepromSize, savedata->mapMode);
 	}
-	savedata->dust = 0;
 	savedata->realisticTiming = realisticTiming;
-	if (end < SIZE_CART_EEPROM) {
-		memset(&savedata->data[end], 0xFF, SIZE_CART_EEPROM - end);
+	if (end < SIZE_CART_EEPROM512) {
+		memset(&savedata->data[end], 0xFF, SIZE_CART_EEPROM512 - end);
 	}
 }
 
@@ -304,10 +313,7 @@ uint8_t GBASavedataReadFlash(struct GBASavedata* savedata, uint16_t address) {
 			}
 		}
 	}
-	if (savedata->dust > 0 && (address >> 12) == savedata->settling) {
-		// Give some overhead for waitstates and the comparison
-		// This estimation can probably be improved
-		savedata->dust -= 5000;
+	if (mTimingIsScheduled(savedata->timing, &savedata->dust) && (address >> 12) == savedata->settling) {
 		return 0x5F;
 	}
 	return savedata->currentBank[address];
@@ -322,7 +328,8 @@ void GBASavedataWriteFlash(struct GBASavedata* savedata, uint16_t address, uint8
 			savedata->currentBank[address] = value;
 			savedata->command = FLASH_COMMAND_NONE;
 			if (savedata->realisticTiming) {
-				savedata->dust = FLASH_PROGRAM_CYCLES;
+				mTimingDeschedule(savedata->timing, &savedata->dust);
+				mTimingSchedule(savedata->timing, &savedata->dust, FLASH_PROGRAM_CYCLES);
 			}
 			break;
 		case FLASH_COMMAND_SWITCH_BANK:
@@ -401,6 +408,19 @@ void GBASavedataWriteFlash(struct GBASavedata* savedata, uint16_t address, uint8
 	}
 }
 
+static void _ensureEeprom(struct GBASavedata* savedata, uint32_t size) {
+	if (size < SIZE_CART_EEPROM512) {
+		return;
+	}
+	if (!savedata->vf || savedata->vf->size(savedata->vf) > SIZE_CART_EEPROM512) {
+		return;
+	}
+	savedata->vf->unmap(savedata->vf, savedata->data, SIZE_CART_EEPROM512);
+	savedata->vf->truncate(savedata->vf, SIZE_CART_EEPROM);
+	savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_EEPROM, savedata->mapMode);
+	memset(&savedata->data[SIZE_CART_EEPROM512], 0xFF, SIZE_CART_EEPROM - SIZE_CART_EEPROM512);
+}
+
 void GBASavedataWriteEEPROM(struct GBASavedata* savedata, uint16_t value, uint32_t writeSize) {
 	switch (savedata->command) {
 	// Read header
@@ -426,13 +446,15 @@ void GBASavedataWriteEEPROM(struct GBASavedata* savedata, uint16_t value, uint32
 		} else if (writeSize == 1) {
 			savedata->command = EEPROM_COMMAND_NULL;
 		} else if ((savedata->writeAddress >> 3) < SIZE_CART_EEPROM) {
+			_ensureEeprom(savedata, savedata->writeAddress >> 3);
 			uint8_t current = savedata->data[savedata->writeAddress >> 3];
 			current &= ~(1 << (0x7 - (savedata->writeAddress & 0x7)));
 			current |= (value & 0x1) << (0x7 - (savedata->writeAddress & 0x7));
 			savedata->dirty |= SAVEDATA_DIRT_NEW;
 			savedata->data[savedata->writeAddress >> 3] = current;
 			if (savedata->realisticTiming) {
-				savedata->dust = EEPROM_SETTLE_CYCLES;
+				mTimingDeschedule(savedata->timing, &savedata->dust);
+				mTimingSchedule(savedata->timing, &savedata->dust, EEPROM_SETTLE_CYCLES);
 			}
 			++savedata->writeAddress;
 		} else {
@@ -456,12 +478,9 @@ void GBASavedataWriteEEPROM(struct GBASavedata* savedata, uint16_t value, uint32
 
 uint16_t GBASavedataReadEEPROM(struct GBASavedata* savedata) {
 	if (savedata->command != EEPROM_COMMAND_READ) {
-		if (!savedata->realisticTiming || savedata->dust <= 0) {
+		if (!savedata->realisticTiming || !mTimingIsScheduled(savedata->timing, &savedata->dust)) {
 			return 1;
 		} else {
-			// Give some overhead for waitstates and the comparison
-			// This estimation can probably be improved
-			--savedata->dust;
 			return 0;
 		}
 	}
@@ -469,6 +488,7 @@ uint16_t GBASavedataReadEEPROM(struct GBASavedata* savedata) {
 	if (savedata->readBitsRemaining < 64) {
 		int step = 63 - savedata->readBitsRemaining;
 		uint32_t address = (savedata->readAddress + step) >> 3;
+		_ensureEeprom(savedata, address);
 		if (address >= SIZE_CART_EEPROM) {
 			mLOG(GBA_SAVE, GAME_ERROR, "Reading beyond end of EEPROM: %08X", address);
 			return 0xFF;
@@ -512,12 +532,18 @@ void GBASavedataSerialize(const struct GBASavedata* savedata, struct GBASerializ
 	GBASerializedSavedataFlags flags = 0;
 	flags = GBASerializedSavedataFlagsSetFlashState(flags, savedata->flashState);
 	flags = GBASerializedSavedataFlagsTestFillFlashBank(flags, savedata->currentBank == &savedata->data[0x10000]);
+
+	if (mTimingIsScheduled(savedata->timing, &savedata->dust)) {
+		STORE_32(savedata->dust.when - mTimingCurrentTime(savedata->timing), 0, &state->savedata.settlingDust);
+		flags = GBASerializedSavedataFlagsFillDustSettling(flags);
+	}
+
 	state->savedata.flags = flags;
-	STORE_32(savedata->readBitsRemaining, 0, &state->savedata.readBitsRemaining);
+	state->savedata.readBitsRemaining = savedata->readBitsRemaining;
 	STORE_32(savedata->readAddress, 0, &state->savedata.readAddress);
 	STORE_32(savedata->writeAddress, 0, &state->savedata.writeAddress);
 	STORE_16(savedata->settling, 0, &state->savedata.settlingSector);
-	STORE_16(savedata->dust, 0, &state->savedata.settlingDust);
+
 }
 
 void GBASavedataDeserialize(struct GBASavedata* savedata, const struct GBASerializedState* state) {
@@ -528,25 +554,37 @@ void GBASavedataDeserialize(struct GBASavedata* savedata, const struct GBASerial
 	savedata->command = state->savedata.command;
 	GBASerializedSavedataFlags flags = state->savedata.flags;
 	savedata->flashState = GBASerializedSavedataFlagsGetFlashState(flags);
-	LOAD_32(savedata->readBitsRemaining, 0, &state->savedata.readBitsRemaining);
+	savedata->readBitsRemaining = state->savedata.readBitsRemaining;
 	LOAD_32(savedata->readAddress, 0, &state->savedata.readAddress);
 	LOAD_32(savedata->writeAddress, 0, &state->savedata.writeAddress);
 	LOAD_16(savedata->settling, 0, &state->savedata.settlingSector);
-	LOAD_16(savedata->dust, 0, &state->savedata.settlingDust);
 
 	if (savedata->type == SAVEDATA_FLASH1M) {
 		_flashSwitchBank(savedata, GBASerializedSavedataFlagsGetFlashBank(flags));
 	}
+
+	if (GBASerializedSavedataFlagsIsDustSettling(flags)) {
+		uint32_t when;
+		LOAD_32(when, 0, &state->savedata.settlingDust);
+		mTimingSchedule(savedata->timing, &savedata->dust, when);
+	}
 }
 
 void _flashSwitchBank(struct GBASavedata* savedata, int bank) {
 	mLOG(GBA_SAVE, DEBUG, "Performing flash bank switch to bank %i", bank);
 	savedata->currentBank = &savedata->data[bank << 16];
 	if (bank > 0 && savedata->type == SAVEDATA_FLASH512) {
+		mLOG(GBA_SAVE, INFO, "Updating flash chip from 512kb to 1Mb");
 		savedata->type = SAVEDATA_FLASH1M;
 		if (savedata->vf) {
-			savedata->vf->truncate(savedata->vf, SIZE_CART_FLASH1M);
-			memset(&savedata->data[SIZE_CART_FLASH512], 0xFF, SIZE_CART_FLASH512);
+			savedata->vf->unmap(savedata->vf, savedata->data, SIZE_CART_FLASH512);
+			if (savedata->vf->size(savedata->vf) == SIZE_CART_FLASH512) {
+				savedata->vf->truncate(savedata->vf, SIZE_CART_FLASH1M);
+				savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_FLASH1M, MAP_WRITE);
+				memset(&savedata->data[SIZE_CART_FLASH512], 0xFF, SIZE_CART_FLASH512);
+			} else {
+				savedata->data = savedata->vf->map(savedata->vf, SIZE_CART_FLASH1M, MAP_WRITE);
+			}
 		}
 	}
 }
@@ -570,7 +608,8 @@ void _flashEraseSector(struct GBASavedata* savedata, uint16_t sectorStart) {
 	}
 	savedata->settling = sectorStart >> 12;
 	if (savedata->realisticTiming) {
-		savedata->dust = FLASH_ERASE_CYCLES;
+		mTimingDeschedule(savedata->timing, &savedata->dust);
+		mTimingSchedule(savedata->timing, &savedata->dust, FLASH_ERASE_CYCLES);
 	}
 	memset(&savedata->currentBank[sectorStart & ~(size - 1)], 0xFF, size);
 }
