@@ -1,4 +1,4 @@
-/* Copyright (c) 2013-2016 Jeffrey Pfau
+/* Copyright (c) 2013-2017 Jeffrey Pfau
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -6,11 +6,8 @@
 #ifndef QGBA_ARCHIVE_INSPECTOR
 #define QGBA_ARCHIVE_INSPECTOR
 
-#include "LibraryModel.h"
-
 #include "ui_ArchiveInspector.h"
 
-struct VDir;
 struct VFile;
 
 namespace QGBA {
@@ -20,15 +17,12 @@ Q_OBJECT
 
 public:
 	ArchiveInspector(const QString& filename, QWidget* parent = nullptr);
-	virtual ~ArchiveInspector();
 
 	VFile* selectedVFile() const;
+	QPair<QString, QString> selectedPath() const;
 
 private:
 	Ui::ArchiveInspector m_ui;
-
-	LibraryModel m_model;
-	VDir* m_dir;
 };
 
 }
