@@ -11,9 +11,7 @@
 #include <QDir>
 #include <QMenu>
 
-extern "C" {
-#include "feature/commandline.h"
-}
+#include <mgba/feature/commandline.h>
 
 using namespace QGBA;
 
@@ -78,25 +76,22 @@ void ConfigOption::setValue(const char* value) {
 }
 
 void ConfigOption::setValue(const QVariant& value) {
-	QPair<QAction*, QVariant> action;
-	foreach (action, m_actions) {
+	for (QPair<QAction*, QVariant>& action : m_actions) {
 		bool signalsEnabled = action.first->blockSignals(true);
 		action.first->setChecked(value == action.second);
 		action.first->blockSignals(signalsEnabled);
 	}
-	std::function<void(const QVariant&)> slot;
-	foreach(slot, m_slots.values()) {
+	for (std::function<void(const QVariant&)>& slot : m_slots.values()) {
 		slot(value);
 	}
 }
 
+QString ConfigController::s_configDir;
+
 ConfigController::ConfigController(QObject* parent)
 	: QObject(parent)
-	, m_opts()
 {
-	char path[PATH_MAX];
-	mCoreConfigDirectory(path, sizeof(path));
-	QString fileName(path);
+	QString fileName = configDir();
 	fileName.append(QDir::separator());
 	fileName.append("qt.ini");
 	m_settings = new QSettings(fileName, QSettings::IniFormat, this);
@@ -112,6 +107,7 @@ ConfigController::ConfigController(QObject* parent)
 	m_opts.logLevel = mLOG_WARN | mLOG_ERROR | mLOG_FATAL;
 	m_opts.rewindEnable = false;
 	m_opts.rewindBufferCapacity = 300;
+	m_opts.rewindSave = true;
 	m_opts.useBios = true;
 	m_opts.suspendScreensaver = true;
 	m_opts.lockAspectRatio = true;
@@ -166,6 +162,10 @@ QString ConfigController::getOption(const char* key) const {
 	return QString(mCoreConfigGetValue(&m_config, key));
 }
 
+QString ConfigController::getOption(const QString& key) const {
+	return getOption(key.toUtf8().constData());
+}
+
 QVariant ConfigController::getQtOption(const QString& key, const QString& group) const {
 	if (!group.isNull()) {
 		m_settings->beginGroup(group);
@@ -271,9 +271,7 @@ void ConfigController::write() {
 void ConfigController::makePortable() {
 	mCoreConfigMakePortable(&m_config);
 
-	char path[PATH_MAX];
-	mCoreConfigDirectory(path, sizeof(path));
-	QString fileName(path);
+	QString fileName(configDir());
 	fileName.append(QDir::separator());
 	fileName.append("qt.ini");
 	QSettings* settings2 = new QSettings(fileName, QSettings::IniFormat, this);
@@ -283,3 +281,12 @@ void ConfigController::makePortable() {
 	delete m_settings;
 	m_settings = settings2;
 }
+
+const QString& ConfigController::configDir() {
+	if (s_configDir.isNull()) {
+		char path[PATH_MAX];
+		mCoreConfigDirectory(path, sizeof(path));
+		s_configDir = QString::fromUtf8(path);
+	}
+	return s_configDir;
+}
