@@ -15,11 +15,9 @@
 
 #include <functional>
 
-extern "C" {
-#include "core/config.h"
-#include "util/configuration.h"
-#include "feature/commandline.h"
-}
+#include <mgba/core/config.h>
+#include <mgba-util/configuration.h>
+#include <mgba/feature/commandline.h>
 
 class QAction;
 class QMenu;
@@ -73,6 +71,7 @@ public:
 	void updateOption(const char* key);
 
 	QString getOption(const char* key) const;
+	QString getOption(const QString& key) const;
 
 	QVariant getQtOption(const QString& key, const QString& group = QString()) const;
 
@@ -86,6 +85,8 @@ public:
 
 	const mCoreConfig* config() { return &m_config; }
 
+	static const QString& configDir();
+
 public slots:
 	void setOption(const char* key, bool value);
 	void setOption(const char* key, int value);
@@ -101,10 +102,11 @@ private:
 	Configuration* defaults() { return &m_config.defaultsTable; }
 
 	mCoreConfig m_config;
-	mCoreOptions m_opts;
+	mCoreOptions m_opts{};
 
 	QMap<QString, ConfigOption*> m_optionSet;
 	QSettings* m_settings;
+	static QString s_configDir;
 };
 
 }
