@@ -5,12 +5,14 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #include "DisplayGL.h"
 
+#if defined(BUILD_GL) || defined(BUILD_GLES2)
+
 #include <QApplication>
 #include <QResizeEvent>
+#include <QTimer>
 
-extern "C" {
-#include "core/core.h"
-#include "core/thread.h"
+#include <mgba/core/core.h>
+#include <mgba/core/thread.h>
 #ifdef BUILD_GL
 #include "platform/opengl/gl.h"
 #endif
@@ -20,16 +22,12 @@ extern "C" {
 #include <epoxy/wgl.h>
 #endif
 #endif
-}
 
 using namespace QGBA;
 
 DisplayGL::DisplayGL(const QGLFormat& format, QWidget* parent)
 	: Display(parent)
-	, m_isDrawing(false)
 	, m_gl(new EmptyGLWidget(format, this))
-	, m_drawThread(nullptr)
-	, m_context(nullptr)
 {
 	m_painter = new PainterGL(format.majorVersion() < 2 ? 1 : m_gl->format().majorVersion(), m_gl);
 	m_gl->setMouseTracking(true);
@@ -69,13 +67,18 @@ void DisplayGL::startDrawing(mCoreThread* thread) {
 	m_gl->context()->doneCurrent();
 	m_gl->context()->moveToThread(m_drawThread);
 	m_painter->moveToThread(m_drawThread);
-	connect(m_drawThread, SIGNAL(started()), m_painter, SLOT(start()));
+	connect(m_drawThread, &QThread::started, m_painter, &PainterGL::start);
 	m_drawThread->start();
 	mCoreSyncSetVideoSync(&m_context->sync, false);
 
 	lockAspectRatio(isAspectRatioLocked());
+	lockIntegerScaling(isIntegerScalingLocked());
 	filter(isFiltered());
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 6, 0))
+	messagePainter()->resize(size(), isAspectRatioLocked(), devicePixelRatioF());
+#else
 	messagePainter()->resize(size(), isAspectRatioLocked(), devicePixelRatio());
+#endif
 	resizePainter();
 }
 
@@ -133,6 +136,13 @@ void DisplayGL::lockAspectRatio(bool lock) {
 	}
 }
 
+void DisplayGL::lockIntegerScaling(bool lock) {
+	Display::lockIntegerScaling(lock);
+	if (m_drawThread) {
+		QMetaObject::invokeMethod(m_painter, "lockIntegerScaling", Q_ARG(bool, lock));
+	}
+}
+
 void DisplayGL::filter(bool filter) {
 	Display::filter(filter);
 	if (m_drawThread) {
@@ -287,6 +297,13 @@ void PainterGL::lockAspectRatio(bool lock) {
 	}
 }
 
+void PainterGL::lockIntegerScaling(bool lock) {
+	m_backend->lockIntegerScaling = lock;
+	if (m_started && !m_active) {
+		forceDraw();
+	}
+}
+
 void PainterGL::filter(bool filter) {
 	m_backend->filter = filter;
 	if (m_started && !m_active) {
@@ -315,6 +332,7 @@ void PainterGL::draw() {
 	if (m_queue.isEmpty() || !mCoreThreadIsActive(m_context)) {
 		return;
 	}
+
 	if (mCoreSyncWaitFrameStart(&m_context->sync) || !m_queue.isEmpty()) {
 		dequeue();
 		mCoreSyncWaitFrameEnd(&m_context->sync);
@@ -322,6 +340,14 @@ void PainterGL::draw() {
 		performDraw();
 		m_painter.end();
 		m_backend->swap(m_backend);
+		if (!m_delayTimer.isValid()) {
+			m_delayTimer.start();
+		} else {
+			while (m_delayTimer.elapsed() < 15) {
+				QThread::usleep(100);
+			}
+			m_delayTimer.restart();
+		}
 	} else {
 		mCoreSyncWaitFrameEnd(&m_context->sync);
 	}
@@ -362,7 +388,11 @@ void PainterGL::unpause() {
 
 void PainterGL::performDraw() {
 	m_painter.beginNativePainting();
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 6, 0))
+	float r = m_gl->devicePixelRatioF();
+#else
 	float r = m_gl->devicePixelRatio();
+#endif
 	m_backend->resized(m_backend, m_size.width() * r, m_size.height() * r);
 	m_backend->drawFrame(m_backend);
 	m_painter.endNativePainting();
@@ -452,3 +482,5 @@ void PainterGL::clearShaders() {
 VideoShader* PainterGL::shaders() {
 	return &m_shader;
 }
+
+#endif
