@@ -6,19 +6,17 @@
 #ifndef QGBA_GDB_CONTROLLER
 #define QGBA_GDB_CONTROLLER
 
-#include <QObject>
+#include "DebuggerController.h"
 
 #ifdef USE_GDB_STUB
 
-extern "C" {
-#include "debugger/gdb-stub.h"
-}
+#include <mgba/internal/debugger/gdb-stub.h>
 
 namespace QGBA {
 
 class GameController;
 
-class GDBController : public QObject {
+class GDBController : public DebuggerController {
 Q_OBJECT
 
 public:
@@ -31,23 +29,19 @@ public:
 public slots:
 	void setPort(ushort port);
 	void setBindAddress(uint32_t bindAddress);
-	void attach();
-	void detach();
 	void listen();
-	void breakInto();
 
 signals:
 	void listening();
 	void listenFailed();
 
 private:
+	virtual void shutdownInternal() override;
+
 	GDBStub m_gdbStub;
-	GameController* m_gameController;
 
-	ushort m_port;
+	ushort m_port = 2345;
 	Address m_bindAddress;
-
-	QMetaObject::Connection m_autoattach;
 };
 
 }
