@@ -12,6 +12,9 @@
 
 #include <QMap>
 
+#include <mgba/internal/gba/gba.h>
+#include <mgba/internal/gba/video.h>
+
 using namespace QGBA;
 
 GIFView::GIFView(QWidget* parent)
@@ -19,14 +22,15 @@ GIFView::GIFView(QWidget* parent)
 {
 	m_ui.setupUi(this);
 
-	connect(m_ui.start, SIGNAL(clicked()), this, SLOT(startRecording()));
-	connect(m_ui.stop, SIGNAL(clicked()), this, SLOT(stopRecording()));
+	connect(m_ui.start, &QAbstractButton::clicked, this, &GIFView::startRecording);
+	connect(m_ui.stop, &QAbstractButton::clicked, this, &GIFView::stopRecording);
 
-	connect(m_ui.selectFile, SIGNAL(clicked()), this, SLOT(selectFile()));
-	connect(m_ui.filename, SIGNAL(textChanged(const QString&)), this, SLOT(setFilename(const QString&)));
+	connect(m_ui.selectFile, &QAbstractButton::clicked, this, &GIFView::selectFile);
+	connect(m_ui.filename, &QLineEdit::textChanged, this, &GIFView::setFilename);
 
-	connect(m_ui.frameskip, SIGNAL(valueChanged(int)), this, SLOT(updateDelay()));
-	connect(m_ui.delayAuto, SIGNAL(clicked(bool)), this, SLOT(updateDelay()));
+	connect(m_ui.frameskip, static_cast<void (QSpinBox::*)(int)>(&QSpinBox::valueChanged),
+	        this, &GIFView::updateDelay);
+	connect(m_ui.delayAuto, &QAbstractButton::clicked, this, &GIFView::updateDelay);
 
 	ImageMagickGIFEncoderInit(&m_encoder);
 }
