@@ -12,13 +12,13 @@
 #include <QGridLayout>
 #include <QSpinBox>
 
-extern "C" {
-#include "gba/io.h"
-}
+#include <mgba/internal/gba/io.h>
+#include <mgba/internal/gba/memory.h>
+
+struct ARMCore;
 
 using namespace QGBA;
 
-
 QList<IOViewer::RegisterDescription> IOViewer::s_registers;
 
 const QList<IOViewer::RegisterDescription>& IOViewer::registerDescriptions() {
@@ -1040,9 +1040,10 @@ IOViewer::IOViewer(GameController* controller, QWidget* parent)
 	const QFont font = QFontDatabase::systemFont(QFontDatabase::FixedFont);
 	m_ui.regValue->setFont(font);
 
-	connect(m_ui.buttonBox, SIGNAL(clicked(QAbstractButton*)), this, SLOT(buttonPressed(QAbstractButton*)));
-	connect(m_ui.buttonBox, SIGNAL(rejected()), this, SLOT(close()));
-	connect(m_ui.regSelect, SIGNAL(currentIndexChanged(int)), this, SLOT(selectRegister()));
+	connect(m_ui.buttonBox, &QDialogButtonBox::clicked, this, &IOViewer::buttonPressed);
+	connect(m_ui.buttonBox, &QDialogButtonBox::rejected, this, &QWidget::close);
+	connect(m_ui.regSelect, &QComboBox::currentTextChanged,
+	        this, static_cast<void (IOViewer::*)()>(&IOViewer::selectRegister));
 
 	m_b[0] = m_ui.b0;
 	m_b[1] = m_ui.b1;
@@ -1062,7 +1063,7 @@ IOViewer::IOViewer(GameController* controller, QWidget* parent)
 	m_b[15] = m_ui.bF;
 
 	for (int i = 0; i < 16; ++i) {
-		connect(m_b[i], SIGNAL(toggled(bool)), this, SLOT(bitFlipped()));
+		connect(m_b[i], &QAbstractButton::toggled, this, &IOViewer::bitFlipped);
 	}
 
 	selectRegister(0);
@@ -1128,8 +1129,8 @@ void IOViewer::selectRegister(unsigned address) {
 				QCheckBox* check = new QCheckBox;
 				check->setEnabled(!ri.readonly);
 				box->addWidget(check, i, 1, Qt::AlignRight);
-				connect(check, SIGNAL(toggled(bool)), m_b[ri.start], SLOT(setChecked(bool)));
-				connect(m_b[ri.start], SIGNAL(toggled(bool)), check, SLOT(setChecked(bool)));
+				connect(check, &QAbstractButton::toggled, m_b[ri.start], &QAbstractButton::setChecked);
+				connect(m_b[ri.start], &QAbstractButton::toggled, check, &QAbstractButton::setChecked);
 			} else if (ri.items.empty()) {
 				QSpinBox* sbox = new QSpinBox;
 				sbox->setEnabled(!ri.readonly);
