@@ -14,10 +14,10 @@ LogController::LogController(int levels, QObject* parent)
 	, m_logLevel(levels)
 {
 	if (this != &s_global) {
-		connect(&s_global, SIGNAL(logPosted(int, int, const QString&)), this, SLOT(postLog(int, int, const QString&)));
-		connect(this, SIGNAL(levelsSet(int)), &s_global, SLOT(setLevels(int)));
-		connect(this, SIGNAL(levelsEnabled(int)), &s_global, SLOT(enableLevels(int)));
-		connect(this, SIGNAL(levelsDisabled(int)), &s_global, SLOT(disableLevels(int)));
+		connect(&s_global, &LogController::logPosted, this, &LogController::postLog);
+		connect(this, &LogController::levelsSet, &s_global, &LogController::setLevels);
+		connect(this, &LogController::levelsEnabled, &s_global, &LogController::enableLevels);
+		connect(this, &LogController::levelsDisabled, &s_global, &LogController::disableLevels);
 	}
 }
 
@@ -72,9 +72,9 @@ QString LogController::toString(int level) {
 }
 
 LogController::Stream::Stream(LogController* controller, int level, int category)
-	: m_log(controller)
-	, m_level(level)
+	: m_level(level)
 	, m_category(category)
+	, m_log(controller)
 {
 }
 
