@@ -11,10 +11,6 @@
 #include <QObject>
 #include <QStringList>
 
-extern "C" {
-#include "gba/gba.h"
-}
-
 namespace QGBA {
 
 class LogController : public QObject {
@@ -64,7 +60,7 @@ private:
 	static LogController s_global;
 };
 
-#define LOG(C, L) (*LogController::global())(_mLOG_CAT_ ## C (), mLOG_ ## L)
+#define LOG(C, L) (*LogController::global())(mLOG_ ## L, _mLOG_CAT_ ## C ())
 
 }
 
