@@ -9,9 +9,7 @@
 #include <QDialog>
 
 #ifdef M_CORE_GB
-extern "C" {
-#include "gb/overrides.h"
-}
+#include <mgba/gb/interface.h>
 #endif
 
 #include "ui_OverrideView.h"
@@ -38,13 +36,19 @@ private slots:
 	void gameStarted(mCoreThread*);
 	void gameStopped();
 
+protected:
+	bool eventFilter(QObject* obj, QEvent* event) override;
+
 private:
 	Ui::OverrideView m_ui;
 
 	GameController* m_controller;
 	ConfigController* m_config;
+	bool m_savePending = false;
 
 #ifdef M_CORE_GB
+	uint32_t m_gbColors[4]{};
+
 	static QList<enum GBModel> s_gbModelList;
 	static QList<enum GBMemoryBankControllerType> s_mbcList;
 #endif
