@@ -1,4 +1,4 @@
-/* Copyright (c) 2013-2015 Jeffrey Pfau
+/* Copyright (c) 2013-2016 Jeffrey Pfau
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -13,12 +13,8 @@
 
 #include <functional>
 
-extern "C" {
-#include "core/thread.h"
-#include "gba/gba.h"
-}
+#include <mgba/core/thread.h>
 
-#include "GDBController.h"
 #include "InputController.h"
 #include "LoadSaveState.h"
 #include "LogController.h"
@@ -27,9 +23,12 @@ struct mArguments;
 namespace QGBA {
 
 class ConfigController;
+class DebuggerConsoleController;
 class Display;
 class GameController;
+class GDBController;
 class GIFView;
+class LibraryController;
 class LogView;
 class ShaderSelector;
 class ShortcutController;
@@ -50,6 +49,8 @@ public:
 
 	void resizeFrame(const QSize& size);
 
+	void updateMultiplayerStatus(bool canOpenAnother) { m_multiWindow->setEnabled(canOpenAnother); }
+
 signals:
 	void startDrawing(mCoreThread*);
 	void shutdown();
@@ -59,9 +60,11 @@ signals:
 
 public slots:
 	void selectROM();
+#ifdef USE_SQLITE3
 	void selectROMInArchive();
+	void addDirToLibrary();
+#endif
 	void selectSave(bool temporary);
-	void selectBIOS();
 	void selectPatch();
 	void enterFullScreen();
 	void exitFullScreen();
@@ -78,17 +81,13 @@ public slots:
 	void exportSharkport();
 
 	void openSettingsWindow();
-	void openOverrideWindow();
-	void openSensorWindow();
-	void openCheatsWindow();
-
-	void openPaletteWindow();
-	void openTileWindow();
-	void openMemoryWindow();
-	void openIOViewer();
-
 	void openAboutScreen();
-	void openROMInfo();
+
+	void startVideoLog();
+
+#ifdef USE_DEBUGGERS
+	void consoleOpen();
+#endif
 
 #ifdef USE_FFMPEG
 	void openVideoWindow();
@@ -143,6 +142,9 @@ private:
 
 	void openView(QWidget* widget);
 
+	template <typename T, typename A> std::function<void()> openTView(A arg);
+	template <typename T> std::function<void()> openTView();
+
 	QAction* addControlledAction(QMenu* menu, QAction* action, const QString& name);
 	QAction* addHiddenAction(QMenu* menu, QAction* action, const QString& name);
 
@@ -160,37 +162,47 @@ private:
 #ifdef M_CORE_GBA
 	QList<QAction*> m_gbaActions;
 #endif
+	QAction* m_multiWindow;
 	QMap<int, QAction*> m_frameSizes;
-	LogController m_log;
+	LogController m_log{0};
 	LogView* m_logView;
-	LoadSaveState* m_stateWindow;
+#ifdef USE_DEBUGGERS
+	DebuggerConsoleController* m_console = nullptr;
+#endif
+	LoadSaveState* m_stateWindow = nullptr;
 	WindowBackground* m_screenWidget;
-	QPixmap m_logo;
+	QPixmap m_logo{":/res/mgba-1024.png"};
 	ConfigController* m_config;
 	InputController m_inputController;
 	QList<QDateTime> m_frameList;
 	QTimer m_fpsTimer;
 	QList<QString> m_mruFiles;
-	QMenu* m_mruMenu;
+	QMenu* m_mruMenu = nullptr;
+	QMenu* m_videoLayers;
+	QMenu* m_audioChannels;
 	ShortcutController* m_shortcutController;
 	ShaderSelector* m_shaderView;
-	bool m_fullscreenOnStart;
+	bool m_fullscreenOnStart = false;
 	QTimer m_focusCheck;
-	bool m_autoresume;
-	bool m_wasOpened;
+	bool m_autoresume = false;
+	bool m_wasOpened = false;
 
 	bool m_hitUnimplementedBiosCall;
 
 #ifdef USE_FFMPEG
-	VideoView* m_videoView;
+	VideoView* m_videoView = nullptr;
 #endif
 
 #ifdef USE_MAGICK
-	GIFView* m_gifView;
+	GIFView* m_gifView = nullptr;
 #endif
 
 #ifdef USE_GDB_STUB
-	GDBController* m_gdbController;
+	GDBController* m_gdbController = nullptr;
+#endif
+
+#ifdef USE_SQLITE3
+	LibraryController* m_libraryView;
 #endif
 };
 
@@ -202,7 +214,9 @@ public:
 
 	void setSizeHint(const QSize& size);
 	virtual QSize sizeHint() const override;
-	void setLockAspectRatio(int width, int height);
+	void setDimensions(int width, int height);
+	void setLockIntegerScaling(bool lock);
+	void setLockAspectRatio(bool lock);
 
 protected:
 	virtual void paintEvent(QPaintEvent*) override;
@@ -211,6 +225,8 @@ private:
 	QSize m_sizeHint;
 	int m_aspectWidth;
 	int m_aspectHeight;
+	bool m_lockAspectRatio;
+	bool m_lockIntegerScaling;
 };
 
 }
