@@ -7,17 +7,10 @@
 
 #include "gl-common.h"
 
-#include "core/core.h"
-#include "core/thread.h"
+#include <mgba/core/core.h>
+#include <mgba/core/thread.h>
 #include "platform/opengl/gl.h"
 
-static void _doViewport(int w, int h, struct VideoBackend* v) {
-	v->resized(v, w, h);
-	v->clear(v);
-	v->swap(v);
-	v->clear(v);
-}
-
 static bool mSDLGLInit(struct mSDLRenderer* renderer);
 static void mSDLGLRunloop(struct mSDLRenderer* renderer, void* user);
 static void mSDLGLDeinit(struct mSDLRenderer* renderer);
@@ -38,12 +31,13 @@ bool mSDLGLInit(struct mSDLRenderer* renderer) {
 	mGLContextCreate(&renderer->gl);
 	renderer->gl.d.user = renderer;
 	renderer->gl.d.lockAspectRatio = renderer->lockAspectRatio;
+	renderer->gl.d.lockIntegerScaling = renderer->lockIntegerScaling;
 	renderer->gl.d.filter = renderer->filter;
 	renderer->gl.d.swap = mSDLGLCommonSwap;
 	renderer->gl.d.init(&renderer->gl.d, 0);
 	renderer->gl.d.setDimensions(&renderer->gl.d, renderer->width, renderer->height);
 
-	_doViewport(renderer->viewportWidth, renderer->viewportHeight, &renderer->gl.d);
+	mSDLGLDoViewport(renderer->viewportWidth, renderer->viewportHeight, &renderer->gl.d);
 	return true;
 }
 
@@ -59,7 +53,7 @@ void mSDLGLRunloop(struct mSDLRenderer* renderer, void* user) {
 			// Event handling can change the size of the screen
 			if (renderer->player.windowUpdated) {
 				SDL_GetWindowSize(renderer->window, &renderer->viewportWidth, &renderer->viewportHeight);
-				_doViewport(renderer->viewportWidth, renderer->viewportHeight, v);
+				mSDLGLDoViewport(renderer->viewportWidth, renderer->viewportHeight, v);
 				renderer->player.windowUpdated = 0;
 			}
 #endif
