@@ -12,18 +12,19 @@
 #include <unistd.h>
 #include <wiiuse/wpad.h>
 
-#include "util/common.h"
+#include <mgba-util/common.h>
 
-#include "core/core.h"
+#include <mgba/core/blip_buf.h>
+#include <mgba/core/core.h>
 #include "feature/gui/gui-runner.h"
-#include "gba/audio.h"
-#include "gba/gba.h"
-#include "gba/input.h"
-#include "util/gui.h"
-#include "util/gui/file-select.h"
-#include "util/gui/font.h"
-#include "util/gui/menu.h"
-#include "util/vfs.h"
+#include <mgba/internal/gba/audio.h>
+#include <mgba/internal/gba/gba.h>
+#include <mgba/internal/gba/input.h>
+#include <mgba-util/gui.h>
+#include <mgba-util/gui/file-select.h>
+#include <mgba-util/gui/font.h>
+#include <mgba-util/gui/menu.h>
+#include <mgba-util/vfs.h>
 
 #define GCN1_INPUT 0x47434E31
 #define GCN2_INPUT 0x47434E32
@@ -189,18 +190,20 @@ static void reconfigureScreen(struct mGUIRunner* runner) {
 	free(framebuffer[0]);
 	free(framebuffer[1]);
 
-	framebuffer[0] = SYS_AllocateFramebuffer(vmode);
-	framebuffer[1] = SYS_AllocateFramebuffer(vmode);
-
 	VIDEO_SetBlack(true);
 	VIDEO_Configure(vmode);
+
+	framebuffer[0] = SYS_AllocateFramebuffer(vmode);
+	framebuffer[1] = SYS_AllocateFramebuffer(vmode);
+	VIDEO_ClearFrameBuffer(vmode, framebuffer[0], COLOR_BLACK);
+	VIDEO_ClearFrameBuffer(vmode, framebuffer[1], COLOR_BLACK);
+
 	VIDEO_SetNextFramebuffer(framebuffer[whichFb]);
 	VIDEO_Flush();
 	VIDEO_WaitVSync();
 	if (vmode->viTVMode & VI_NON_INTERLACE) {
 		VIDEO_WaitVSync();
 	}
-	VIDEO_SetBlack(false);
 	GX_SetViewport(0, 0, vmode->fbWidth, vmode->efbHeight, 0, 1);
 
 	f32 yscale = GX_GetYScaleFactor(vmode->efbHeight, vmode->xfbHeight);
@@ -233,6 +236,9 @@ static void reconfigureScreen(struct mGUIRunner* runner) {
 
 int main(int argc, char* argv[]) {
 	VIDEO_Init();
+	VIDEO_SetBlack(true);
+	VIDEO_Flush();
+	VIDEO_WaitVSync();
 	PAD_Init();
 	WPAD_Init();
 	WPAD_SetDataFormat(0, WPAD_FMT_BTNS_ACC_IR);
@@ -472,6 +478,10 @@ int main(int argc, char* argv[]) {
 	mGUIInit(&runner, "wii");
 	reconfigureScreen(&runner);
 
+	// Make sure screen is properly initialized by drawing a blank frame
+	_drawStart();
+	_drawEnd();
+
 	_mapKey(&runner.params.keyMap, GCN1_INPUT, PAD_BUTTON_A, GUI_INPUT_SELECT);
 	_mapKey(&runner.params.keyMap, GCN1_INPUT, PAD_BUTTON_B, GUI_INPUT_BACK);
 	_mapKey(&runner.params.keyMap, GCN1_INPUT, PAD_TRIGGER_Z, GUI_INPUT_CANCEL);
@@ -507,6 +517,9 @@ int main(int argc, char* argv[]) {
 	} else {
 		mGUIRunloop(&runner);
 	}
+	VIDEO_SetBlack(true);
+	VIDEO_Flush();
+	VIDEO_WaitVSync();
 	mGUIDeinit(&runner);
 
 	free(fifo);
@@ -533,6 +546,8 @@ static void _audioDMA(void) {
 }
 
 static void _drawStart(void) {
+	VIDEO_SetBlack(false);
+
 	u32 level = 0;
 	_CPU_ISR_Disable(level);
 	if (referenceRetraceCount > retraceCount) {
@@ -550,12 +565,11 @@ static void _drawStart(void) {
 }
 
 static void _drawEnd(void) {
-	whichFb = !whichFb;
-
 	GX_CopyDisp(framebuffer[whichFb], GX_TRUE);
 	GX_DrawDone();
 	VIDEO_SetNextFramebuffer(framebuffer[whichFb]);
 	VIDEO_Flush();
+	whichFb = !whichFb;
 
 	u32 level = 0;
 	_CPU_ISR_Disable(level);
@@ -640,8 +654,8 @@ void _guiPrepare(void) {
 }
 
 void _setup(struct mGUIRunner* runner) {
-	runner->core->setRotation(runner->core, &rotation);
-	runner->core->setRumble(runner->core, &rumble);
+	runner->core->setPeripheral(runner->core, mPERIPH_ROTATION, &rotation);
+	runner->core->setPeripheral(runner->core, mPERIPH_RUMBLE, &rumble);
 
 	_mapKey(&runner->core->inputMap, GCN1_INPUT, PAD_BUTTON_A, GBA_KEY_A);
 	_mapKey(&runner->core->inputMap, GCN1_INPUT, PAD_BUTTON_B, GBA_KEY_B);
@@ -700,6 +714,9 @@ void _gameUnloaded(struct mGUIRunner* runner) {
 	UNUSED(runner);
 	AUDIO_StopDMA();
 	frameLimiter = true;
+	VIDEO_SetBlack(true);
+	VIDEO_Flush();
+	VIDEO_WaitVSync();
 }
 
 void _gameLoaded(struct mGUIRunner* runner) {
@@ -720,6 +737,7 @@ void _gameLoaded(struct mGUIRunner* runner) {
 
 void _unpaused(struct mGUIRunner* runner) {
 	u32 level = 0;
+	VIDEO_WaitVSync();
 	_CPU_ISR_Disable(level);
 	referenceRetraceCount = retraceCount;
 	_CPU_ISR_Restore(level);
