@@ -3,11 +3,11 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "file-select.h"
+#include <mgba-util/gui/file-select.h>
 
-#include "util/gui/font.h"
-#include "util/gui/menu.h"
-#include "util/vfs.h"
+#include <mgba-util/gui/font.h>
+#include <mgba-util/gui/menu.h>
+#include <mgba-util/vfs.h>
 
 #include <stdlib.h>
 
