@@ -3,11 +3,11 @@
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-#include "util/patch-ips.h"
+#include <mgba-util/patch/ips.h>
 
-#include "util/crc32.h"
-#include "util/patch.h"
-#include "util/vfs.h"
+#include <mgba-util/crc32.h>
+#include <mgba-util/patch.h>
+#include <mgba-util/vfs.h>
 
 enum {
 	IN_CHECKSUM = -12,
@@ -87,6 +87,9 @@ bool _UPSApplyPatch(struct Patch* patch, const void* in, size_t inSize, void* ou
 			if (patch->vf->read(patch->vf, &byte, 1) != 1) {
 				return false;
 			}
+			if (offset >= outSize) {
+				return false;
+			}
 			buf[offset] ^= byte;
 			++offset;
 			if (!byte) {
